import tkinter as tk
from tkinter.filedialog import asksaveasfile, askopenfilename
import xml.etree.ElementTree as ET
import xml.dom.minidom
import copy
from PyTASTileSystem import *


square_size = 20
canvas_size = square_size * 30

color_list     = ["white", "green", "red", "blue", "yellow", "pink", "aqua", "lime", "olive",
                  "orange", "gold", "grey", "purple", "maroon", "turquoise", "black"]
seed_color     = 'green'
coop_color     = 'orange'
non_coop_color = 'blue'

neighbor_locs = [(1, 0), (-1,0), (0, 1), (0, -1)]
OFFSET_TO_DIR = {(1,0): E, (0,1): N, (-1,0): W, (0,-1): S}
DIR_TO_OFFSET = {E: (1,0), N: (0,1), W: (-1,0), S: (0,-1)}

INPUT_MARKERS =  {N: 'v', E: '<', S: '^', W: '>'}
OUTPUT_MARKERS = {N: '^', E: '>', S: 'v', W: '<'}

LABEL_OFFSETS = {N: (int(square_size / 2), 5),
                 E: (square_size - 5, int(square_size/2)),
                 S: (int(square_size/2), square_size - 5),
                 W: (5, int(square_size / 2))}

help_string = """
This program is used to design paths and assemblies that
will self-assemble in the abstract Tile Assembly Model.

Left-click: select a location on a path. The first will
denote the seed tile, and subsequent will denote locations
on the path(s). Left-clicking on a location already on a path
will cause it to cycle through each possible set of tile(s)
which could serve as input to that location, with the
temperature paraemter set to 2.

Right-click: erase a location from an assembly. This will
fail if a location has children (i.e. one or more locations
which use that location as an input).

Mouse-over: when moving the cursor over a location in the
assembly, the coordinates of the input(s) to that location
will be shown followed by '->' then the coordinates of the
location.

Colors:
     seed:      green
     1 input:   blue
     2 inputs:  orange
"""

class DrawnAssemblyLocation():
    def __init__(self, x, y, canvas_objects, parent_list, xml_node=None):
        if xml_node is None:
            self.x = x
            self.y = y
            self.canvas_objects = canvas_objects
            self.parent_list = sorted(parent_list)
        else:
            self.x = None
            self.y = None
            self.canvas_objects = None
            self.parent_list = None
            for elt_child in xml_node:
                if elt_child.tag == 'x':
                    self.x = int(elt_child.text)
                if elt_child.tag == 'y':
                    self.y = int(elt_child.text)
                if elt_child.tag == 'parent_list':
                    self.parent_list = eval(elt_child.text)
            if (self.x is None) or (self.y is None) or (self.parent_list is None):
                print(f'ERROR: Invalid XML node for assembly location.')

    def __str__(self):
        parents = ''
        for parent in self.parent_list:
            parents += f'{(parent[0], parent[1])}'
        ret_str = f'{(self.x, self.y)}'
        if len(parents) > 0:
            ret_str = parents + '->' + ret_str
        return ret_str

    def toXML(self, xml_parent=None):
        elt_dal = ET.SubElement(xml_parent, 'DrawnAssemblyLocation')

        elt_x = ET.SubElement(elt_dal, 'x')
        elt_x.text = str(self.x)

        elt_y = ET.SubElement(elt_dal, 'y')
        elt_y.text = str(self.y)

        elt_parent_list = ET.SubElement(elt_dal, 'parent_list')
        elt_parent_list.text = str(self.parent_list)
        
        return xml_parent

    def get_tile_type(self, color, children_list, output_all=False):
        #print(f'getting tile type for {(self.x, self.y)} with parent_list {self.parent_list} and children_list {children_list}')

        input_dirs = []
        for parent in self.parent_list:
            offset = (parent[0] - self.x, parent[1] - self.y)
            input_dir = OFFSET_TO_DIR[offset]
            input_dirs.append(input_dir)


        output_dirs = {}
        for child in children_list:
            child_coords = child[0]
            offset = (child_coords[0] - self.x, child_coords[1] - self.y)
            output_dir = OFFSET_TO_DIR[offset]
            output_dirs[output_dir] = child[1]     
        if output_all:
            for d in Dirs:
                if (d not in input_dirs) and (d not in output_dirs):
                    output_dirs[d] = 1
            #output_dirs = {output_dir: 1 for output_dir in Dirs if output_dir not in input_dirs}

        tile_type = PyTASTileType()
        for d in [N,E,S,W]:
            strength = 0
            label = ''
            if d in input_dirs:
                strength = 2 if len(input_dirs) == 1 else 1
                input_offset = DIR_TO_OFFSET[d]
                label = f'{INPUT_MARKERS[d]}{(self.x + input_offset[0], self.y + input_offset[1])}'
            elif d in output_dirs:
                strength = output_dirs[d]
                label = f'{OUTPUT_MARKERS[d]}{(self.x, self.y)}'
            tile_type.sides[d] = Glue(strength, label.replace(' ',''))
        tile_type.name = f'{(self.x, self.y)}'.replace(' ','')
        tile_type.label = tile_type.name.replace(' ','')
        tile_type.color = color
        return tile_type

class DrawnAssembly():
    def __init__(self):
        self.output_all = tk.BooleanVar()
        self.Clear()

    def Clear(self):
        self.seed_location = None
        self.output_all.set(False)
        self.coordinates = {}

    def WriteToFile(self, file_name):
        xml_root = ET.Element('DrawnAssembly')

        elt_oad = ET.SubElement(xml_root, 'output_all_directions')
        elt_oad.text = str(self.output_all.get())
        
        if not self.seed_location is None:
            elt_seed = ET.SubElement(xml_root, 'seed_location')
            elt_seed.text = f'{(self.seed_location.x, self.seed_location.y)}'

        if len(self.coordinates) > 0:
            elt_assembly = ET.SubElement(xml_root, 'assembly')
            count = 0
            for coordinate in self.coordinates:
                self.coordinates[coordinate].toXML(elt_assembly)
                count += 1

        xml_string = ET.tostring(xml_root)
        xml_dom = xml.dom.minidom.parseString(xml_string)
        
        with open(file_name, "w") as f: 
            f.write(xml_dom.toprettyxml())
            print(f'Output {count} assembly locations to {file_name}')
        return True

    def ReadFromFile(self, file_name):
        self.Clear()
        xml_tree = ET.parse(file_name)
        xml_root = xml_tree.getroot()
        if not xml_root.tag == 'DrawnAssembly':
            print("Invalid XML DrawnAssembly file")
            return False
        else:
            seed_coords = None
            for child in xml_root:
                if child.tag == 'output_all_directions':
                    if child.text == 'True':
                        self.output_all.set(True)
                    else:
                        self.output_all.set(False)
                elif child.tag == 'seed_location':
                    seed_coords = eval(child.text)
                elif child.tag == 'assembly':
                    for elt_type in child:
                        dal = DrawnAssemblyLocation(x=None, y=None, canvas_objects=None, parent_list=None, xml_node=elt_type)
                        self.coordinates[(dal.x, dal.y)] = dal
            if not seed_coords in self.coordinates:
                print(f'ERROR: seed coordinates {seed_coords} not found in assembly.')
            else:
                self.seed_location = self.coordinates[seed_coords]
            print(f'Loaded assembly with seed coordinates {seed_coords} and {len(self.coordinates)} locations.')
        
class Frame(tk.Frame):
    def __init__(self, master=None):
        super().__init__(master)
        self.master = master
        self.assembly = DrawnAssembly()
        self.tile_system_base_file_name = None

        self.canvas = tk.Canvas(self, width=canvas_size, height=canvas_size)
        
        self.canvas.bind('<Button-1>', self.left_click)
        self.canvas.bind('<B1-Motion>', self.left_click)

        self.canvas.bind('<Button-3>', self.right_click)
        self.canvas.bind('<B3-Motion>', self.right_click)

        self.canvas.bind("<Motion>", self.mouse_moved)
        
        root.bind('<p>', self.print_message)
        self.canvas.pack()

        self.label = tk.Label(self)
        self.label.place(x=200, y=20, anchor="nw")

    def menubar(self, root):
        menubar = tk.Menu(root)

        filemenu = tk.Menu(menubar, tearoff=0)
        
        filemenu.add_command(label="New", command=self.on_new, accelerator='Ctrl+N')
        filemenu.add_separator()
        filemenu.add_command(label="Open...", command=self.on_open, accelerator='Ctrl+O')
        filemenu.add_separator()
        filemenu.add_command(label="Save", command=self.on_save, accelerator='Ctrl+S')
        filemenu.add_command(label="Save as...", command=self.on_save_as)
        filemenu.add_separator()
        filemenu.add_command(label="Save aTAM system", command=self.on_save_aTAM, accelerator='Ctrl+A')
        filemenu.add_command(label="Save aTAM system as...", command=self.on_save_aTAM_as)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", command=self.on_exit, accelerator='Ctrl+X')
        filemenu.bind_all("<Control-n>", self.on_new)
        filemenu.bind_all("<Control-o>", self.on_open)
        filemenu.bind_all("<Control-s>", self.on_save)
        filemenu.bind_all("<Control-a>", self.on_save_aTAM)
        filemenu.bind_all("<Control-x>", self.on_exit)
        menubar.add_cascade(label="File", menu=filemenu)

        settingsmenu = tk.Menu(menubar, tearoff=0)
        settingsmenu.add_checkbutton(label="Output all directions", variable=self.assembly.output_all)
        menubar.add_cascade(label="Settings", menu=settingsmenu)

        helpmenu = tk.Menu(menubar, tearoff=0)
        helpmenu.add_checkbutton(label="Instructions for use", command=self.on_help, accelerator='Ctrl+H')
        helpmenu.bind_all("<Control-h>", self.on_help)
        menubar.add_cascade(label="Help", menu=helpmenu)

        return menubar
        
    def on_help(self, args=None):
        tk.messagebox.showinfo("Instructions for use", help_string)

    def on_exit(self, args=None):
        self.master.destroy()

    def on_new(self, args=None):
        ret = tk.messagebox.askyesno(title='Discard current assembly?',
                                     message='Are you sure you want to discard your current assembly and start a new one?')
        if ret:
            self.assembly.Clear()
            self.tile_system_base_file_name = None
            self.canvas.delete('all')

    def on_open(self, args=None):
        file_path = askopenfilename(title="Select a File", filetypes=[("XML files", "*.xml")])
        if file_path:
            self.assembly.Clear()
            self.tile_system_base_file_name = None
            self.canvas.delete('all')
            
            self.assembly.ReadFromFile(file_path)        
            for assembly_coord in self.assembly.coordinates:
                assembly_location = self.assembly.coordinates[assembly_coord]
                (x, y) = assembly_coord
                canvas_objects = self.create_labeled_rectangle(x, y, assembly_location.parent_list)
                assembly_location.canvas_objects = canvas_objects

    def on_save_as(self):
        file_dialog_ret = asksaveasfile(initialfile = 'tile_paths', defaultextension=".xml",filetypes=[("XML","*.xml")])
        if file_dialog_ret is None:
            return
        self.tile_system_base_file_name = file_dialog_ret.name
        if self.tile_system_base_file_name[-4:] == '.xml':
            self.tile_system_base_file_name = self.tile_system_base_file_name[:-4]
        self.on_save()

    def on_save(self, args=None):
        if self.tile_system_base_file_name is None:
            self.on_save_as()
            return
        self.assembly.WriteToFile(self.tile_system_base_file_name + '.xml')

    def on_save_aTAM_as(self):
        file_dialog_ret = asksaveasfile(initialfile = 'tile_paths', defaultextension=".tdp",filetypes=[("TDP","*.tdp")])
        if file_dialog_ret is None:
            return
        self.tile_system_base_file_name = file_dialog_ret.name
        if self.tile_system_base_file_name[-4:] == '.tdp':
            self.tile_system_base_file_name = self.tile_system_base_file_name[:-4]
        self.on_save_aTAM()

    def on_save_aTAM(self, args=None):
        if self.tile_system_base_file_name is None:
            self.on_save_aTAM_as()
            return
        
        do_output_all = False
        if self.assembly.output_all.get():
            do_output_all = True

        seed_type_name = None
        seed_coords = None
        tile_set_dict = {}
        for assembly_coord in self.assembly.coordinates:
            assembly_location = self.assembly.coordinates[assembly_coord]
            children_list = self.get_children(assembly_coord[0], assembly_coord[1])
            color = self.get_color(assembly_location.parent_list)
            tile_type = assembly_location.get_tile_type(color, children_list, output_all=do_output_all)
            tile_set_dict[tile_type.name] = tile_type
            if len(assembly_location.parent_list) == 0:
                seed_type_name = tile_type.name
                seed_coords = assembly_coord

        tile_system = PyTASTileSystem()
        tile_system.seedPatch.append((seed_type_name, seed_coords))
        tile_system.tileTypes = tile_set_dict
        tile_system.SaveToFile(base_file_name=self.tile_system_base_file_name)
        print(f"Tile system saved to tds and tdp files with base names '{self.tile_system_base_file_name}'")        
        
    def get_coords(self, event):
        x = int(event.x / square_size)
        y = int(canvas_size / square_size) - int(event.y / square_size) - 1
        return (x, y)
    
    def get_pixels(self, x, y):
        x_p = int(x * square_size)
        y_p = canvas_size - int(y * square_size) - square_size
        return (x_p, y_p)

    def format_tag(self, x, y):
        return f'{(x,y)}'

    def get_adjacent(self, x, y):
        neighbors = []
        for (n_x, n_y) in neighbor_locs:
            neighbor = (x + n_x, y + n_y)
            if neighbor in self.assembly.coordinates:
                neighbors.append(neighbor)
        return sorted(neighbors)

    def get_children(self, x, y):
        children_list = []
        for coords in self.assembly.coordinates:
            assembly_location = self.assembly.coordinates[coords]
            for parent in assembly_location.parent_list:
                if parent == (x, y):
                    strength_between = 2 if len(assembly_location.parent_list) == 1 else 1
                    children_list.append((coords, strength_between))
        return children_list

    def get_color(self, parent_list):
        color = seed_color if len(parent_list) == 0 else non_coop_color if len(parent_list) == 1 else coop_color
        return color

    def delete_location(self, x, y):
        assembly_location = self.assembly.coordinates[(x, y)]
        for canvas_object in assembly_location.canvas_objects:
            self.canvas.delete(canvas_object)
        del self.assembly.coordinates[(x,y)]
        
    def create_labeled_rectangle(self, x, y, parent_list):
        (x_p, y_p) = self.get_pixels(x, y)
        color = self.get_color(parent_list)
        rectangle = self.canvas.create_rectangle(x_p, y_p,
                                                 x_p + square_size, y_p + square_size,
                                                 fill=color, outline='white')
        object_list = [rectangle]
        for parent in parent_list:
            parent_offset = (parent[0] - x, parent[1] - y)
            parent_dir = OFFSET_TO_DIR[parent_offset]
            marker = INPUT_MARKERS[parent_dir]
            label_offset = LABEL_OFFSETS[parent_dir]
            label = self.canvas.create_text((x_p + label_offset[0], y_p + label_offset[1]), text=marker, fill='white')
            object_list.append(label)

        return object_list
        
    def left_click(self, event):
        (x, y) = self.get_coords(event)
        
        if not (x, y) in self.assembly.coordinates:
            neighbors = self.get_adjacent(x, y)
            if len(neighbors) > 2:
                neighbors = neighbors[0:2]
                
            if len(self.assembly.coordinates) < 1 or len(neighbors) > 0:
                canvas_objects = self.create_labeled_rectangle(x, y, neighbors)
                self.assembly.coordinates[(x, y)] = DrawnAssemblyLocation(x, y, canvas_objects, neighbors)
                if len(neighbors) == 0:
                    self.assembly.seed_location = self.assembly.coordinates[(x, y)]
            else:
                print(f"ERROR: Can't add a location without a neighbor.")
        elif len(self.assembly.coordinates[(x, y)].parent_list) == 0:
            return
        else:
            all_neighbors = self.get_adjacent(x, y)
            assembly_location = self.assembly.coordinates[(x, y)]
            parents = copy.deepcopy(assembly_location.parent_list)

            # Delete the current objects at this location (after saving the parent_list for comparison to pick a new set
            self.delete_location(x, y)

            # Don't consider neighbors that are children of this one to be its parent(s)
            neighbors = []
            for neighbor in all_neighbors:
                if not (x, y) in self.assembly.coordinates[neighbor].parent_list:
                    neighbors.append(neighbor)

            input_neighbor_sets = [[(neighbor)] for neighbor in neighbors]
            for neighbor1 in neighbors:
                for neighbor2 in neighbors:
                    sorted_neighbors = [neighbor1, neighbor2] if neighbor1 < neighbor2 else [neighbor2, neighbor1]
                    if (not neighbor1 == neighbor2) and (not sorted_neighbors in input_neighbor_sets):
                        input_neighbor_sets.append(sorted_neighbors)
            #print(f'input_neighbor_sets: {input_neighbor_sets}')
            b_found = False
            new_parents = input_neighbor_sets[0]
            for input_neighbor_set in input_neighbor_sets:
                if input_neighbor_set == parents:
                    b_found = True
                elif b_found == True:
                    new_parents = input_neighbor_set
                    break

            canvas_objects = self.create_labeled_rectangle(x, y, new_parents)
            self.assembly.coordinates[(x, y)] = DrawnAssemblyLocation(x, y, canvas_objects, new_parents)


        self.mouse_moved(event)

    def right_click(self, event):
        (x, y) = self.get_coords(event)
        if (x, y) in self.assembly.coordinates:
            children = self.get_children(x, y)
            if len(children) == 0:
                self.delete_location(x, y)
                self.label.config(text=f'{(x,y)}')
            else:
                print('you cannot die... you have kids')

    def mouse_moved(self, event):
        (x, y) = self.get_coords(event)
        self.label.config(text=f'{(x,y)}')
        self.label.place(x=event.x+10, y=event.y+10, anchor="nw")
##        if (x, y) in self.assembly.coordinates:
##            assembly_loc = self.assembly.coordinates[(x, y)]
##            self.label.config(text=assembly_loc)
##            self.label.place(x=event.x+10, y=event.y+10, anchor="nw")
##        else:
##            self.label.config(text=f'{(x,y)}')
##            self.label.place(x=event.x+10, y=event.y+10, anchor="nw")      

    def print_message(self, event):
        print(f'Path:')
        for coordinate in self.assembly.coordinates:
            assembly_loc = self.assembly.coordinates[coordinate]
            print(assembly_loc)


root = tk.Tk()
root.title('aTAM Path Assembly Drawer')

frame = Frame(root)
frame.pack()

menubar = frame.menubar(root)
root.configure(menu=menubar)

root.mainloop()
