from PyTASTileTypeLight import *
import math
import copy
import svgwrite
from svgwrite import cm, mm, px
from reportlab.pdfgen import canvas
import xml.etree.ElementTree as ET
import xml.dom.minidom
from scipy.spatial.distance import hamming
import random
from tabulate import tabulate
from SlatType import TASSlatType
from SlatType import TASSlat
from SlatSystem import SlatSystem
import scadnano as sc


####################
# Colors
color_list = ["white", "green", "red", "blue", "yellow", "pink", "aqua", "lime", "olive", "orange", "gold", "grey", "purple", "maroon", "turquoise", "black"]
# The following colors aren't in the standard wxPython database so the listed conversions will be made for viewing in SlatSimulator
wxPython_color_mapping = {'aqua': 'AQUAMARINE', 'lime': 'LIME GREEN', 'olive': 'KHAKI'}

DIR_TO_STR = {N: 'N', E: 'E', S: 'S', W: 'W'}

def pretty_print_list(results_list, results_labels):
    print(f"\n{tabulate(results_list, headers=results_labels, numalign='center', stralign='center')}\n")

def pretty_print_dict(results_dict, label_mapping_dict=None):
    if label_mapping_dict is None:
        results_list = [list(results_dict.values())]
        results_labels = list(results_dict.keys())
    else:
        results_list = []
        results_labels = []
        for label in label_mapping_dict:
            results_labels.append(label)
            results_list.append(results_dict[label_mapping_dict[label]])
    pretty_print_list([results_list], results_labels)


####################
# DNA functions

bases = ['A', 'C', 'G', 'T']
complementary_bases = {'A': 'T', 'a': 't', 'C': 'G', 'c': 'g', 'G': 'C', 'g': 'c', 'T': 'A', 't': 'a'}

def get_random_DNA_sequence(length, b_as_string=False):
    seq = []
    str_seq = ''
    for l in range(length):
        base = bases[random.randrange(4)]
        seq.append(base)
        str_seq = str_seq + base
    if b_as_string:
        return str_seq
    else:
        return seq

def get_complementary_DNA_seq(seq, b_reverse=False):
    ret_seq = []
    ret_seq_str = ''
    for i in range(len(seq)):
        if type(seq) == list:
            ret_seq.append(complementary_bases[seq[i]])
        else:
            ret_seq_str += complementary_bases[seq[i]]
    if len(ret_seq) > 0:
        if b_reverse:
            return ret_seq[::-1]
        else:
            return ret_seq
    else:
        if b_reverse:
            return ret_seq_str[::-1]
        else:
            return ret_seq_str

def DNA_seqs_complement(l1, l2):
    f = len(l1)
    if len(l1) != len(l2):
        return False
    for i in range(len(l1)):
        if l1[i] != complementary_bases[l2[-(i+1)]]:
            return False
    return True

def read_staples_and_handles(sequences_file):
    staple_sequences_dict = {}
    staple_sequences = []
    handle_sequences = []
    with open(sequences_file, 'r') as sf:
        lines = sf.readlines()
    b_in_staples = False
    b_in_handles = False
    for line in lines:
        line = line.strip()
        if len(line) > 0:
            if line == 'Staples':
                b_in_staples = True
            elif b_in_staples:
                staple_sequences.append(line.strip())
            elif line == 'Handles':
                b_in_handles = True
            elif b_in_handles:
                handle_sequences.append(line.strip())
    staple_sequences_dict[len(staple_sequences)] = staple_sequences
    return (staple_sequences_dict, handle_sequences)

def map_domains_to_sequences(domain_list, handle_list):
    domain_to_handle_mapping = None
    if len(domain_list) <= len(handle_list):
        domain_to_handle_mapping = dict()
        count = 0
        for domain in domain_list:
            domain_to_handle_mapping[domain] = handle_list[count]
            domain_to_handle_mapping[f'{domain}*'] = get_complementary_DNA_seq(handle_list[count], b_reverse=True)
            count += 1
    else:
        print(f'ERROR: There are not enough handles to map uniquely to domains.')
    return domain_to_handle_mapping

def get_slat_types_with_sequences(tas_slat_types_dict, domain_to_handle_map, staple_sequences_list_dict=None, handle_linker=None):
    slat_types_with_sequences_dict = copy.deepcopy(tas_slat_types_dict)
    staple_plus_handle_count_dict = dict()
    staple_count_dict = dict()
    for slat_type_name in slat_types_with_sequences_dict:
        slat_type = slat_types_with_sequences_dict[slat_type_name]
        slat_length = slat_type.getLength()
        staple_sequences_list = None if staple_sequences_list_dict is None else staple_sequences_list_dict[slat_length]
        loc_count = 0
        for b in slat_type.blocks:
            staple_sequence = '' if staple_sequences_list is None else staple_sequences_list[loc_count]
            linker_sequence = '' if handle_linker is None else handle_linker
            if b in slat_type.domains:
                staple_plus_handle = f'{staple_sequence}{linker_sequence}{domain_to_handle_map[slat_type.domains[b][0][0]]}'
                if staple_plus_handle in staple_plus_handle_count_dict:
                    staple_plus_handle_count_dict[staple_plus_handle] += 1
                else:
                    staple_plus_handle_count_dict[staple_plus_handle] = 1
                domain = copy.deepcopy(slat_type.domains[b][0])
                new_domain = tuple([staple_plus_handle] + list(domain[1:]))
                #new_domain = tuple([domain_to_handle_map[domain[0]]] + list(domain[1:]))
                slat_type.domains[b][0] = new_domain
            elif not staple_sequences_list is None:
                if staple_sequence in staple_count_dict:
                    staple_count_dict[staple_sequence] += 1
                else:
                    staple_count_dict[staple_sequence] = 1
                direction = 'U' if slat_type.IsHorizontal() else 'D'
                slat_type.domains[b] = [(staple_sequence, direction, 0)]
            loc_count += 1
    return (slat_types_with_sequences_dict, staple_plus_handle_count_dict, staple_count_dict)

def write_domain_to_sequence_mapping_to_csv(domain_list, domain_to_handle_mapping, csv_file):
    csv_file.write(f'Domain set:\n')
    csv_file.write(f'Number,Domain,Seq,CompSeqRev\n')
    count = 1
    for domain in domain_list:
        mapping = ''
        if not domain_to_handle_mapping is None:
            mapping = f',{domain_to_handle_mapping[domain]},{domain_to_handle_mapping[domain + "*"]}'
        csv_file.write(f'{count},{domain}{mapping}\n')
        count += 1

def write_slat_domains_to_csv(slat_type_dict, csv_file):
    greatest_length = 0
    for slat_type_name in slat_type_dict:
        slat_type = slat_type_dict[slat_type_name]
        length = len(slat_type.blocks)
        if length > greatest_length:
            greatest_length = length

    csv_file.write('Name,')
    for i in range(greatest_length):
        csv_file.write(f'loc{i},')
    csv_file.write('\n')
    for slat_type_name in slat_type_dict:
        slat_type = slat_type_dict[slat_type_name]
        csv_file.write(f'{slat_type_name}')
        for b in slat_type.blocks:
            if b in slat_type.domains:
                domain = slat_type.domains[b][0]
                #print(f'domain: {domain}')
                domain_label = slat_type.domains[b][0][0]
                csv_file.write(f',{domain_label}')
            else:
                csv_file.write(',')
        csv_file.write('\n')

def apply_sequence_mappings_to_system(slat_system, domain_to_handle_mapping, staple_sequences_dict, handle_linker):
    (slat_type_dict_with_handle_seqs, handle_count_dict, staple_count_dict) = get_slat_types_with_sequences(
        slat_system.PolyominoType_dict, domain_to_handle_mapping, None, None)

    (slat_type_dict_with_handle_and_staple_seqs, staple_plus_handle_count_dict,
     staple_count_dict) = get_slat_types_with_sequences(slat_system.PolyominoType_dict, domain_to_handle_mapping,
                                                        staple_sequences_dict, handle_linker)
    seq_slat_system = copy.deepcopy(slat_system)
    seq_slat_system.SwitchToTypeDict(slat_type_dict_with_handle_seqs)
    seq_slat_system.b_use_DNA_seqs = True

    return (seq_slat_system, slat_type_dict_with_handle_seqs, slat_type_dict_with_handle_and_staple_seqs, staple_count_dict, staple_plus_handle_count_dict)

def write_slat_sequence_file(handle_file_name, slat_system, domain_list, domain_to_handle_mapping, slat_type_dict_with_handle_seqs,
                             slat_type_dict_with_handle_and_staple_seqs, staple_count_dict, staple_plus_handle_count_dict):
    with open(handle_file_name, 'w') as hf:
        write_domain_to_sequence_mapping_to_csv(domain_list, domain_to_handle_mapping, hf)

        hf.write(f'\n\nSlat types:\n')
        write_slat_domains_to_csv(slat_system.PolyominoType_dict, hf)

        hf.write(f'\nSlat types with handle DNA:\n')
        write_slat_domains_to_csv(slat_type_dict_with_handle_seqs, hf)

        hf.write(f'\nSlat types with staple and handle DNA:\n')
        write_slat_domains_to_csv(slat_type_dict_with_handle_and_staple_seqs, hf)

        sorted_staples = sorted(staple_count_dict.items(), key=lambda x: x[1])
        hf.write(f'\nStaples ({len(sorted_staples)}):\n')
        hf.write('Number, Count,Sequence\n')
        num = 1
        for staple in sorted_staples:
            hf.write(f'{num},{staple[1]},{staple[0]}\n')
            num += 1

        sorted_staples_plus_handles = sorted(staple_plus_handle_count_dict.items(), key=lambda x: x[1])
        hf.write(f'\nStaples with handles ({len(sorted_staples_plus_handles)}):\n')
        hf.write('Number, Count,Sequence\n')
        num = 1
        for staple_with_handle in sorted_staples_plus_handles:
            hf.write(f'{num},{staple_with_handle[1]},{staple_with_handle[0]}\n')
            num += 1

        print(f'Output sequences to {handle_file_name}')

def check_scadnano_design_for_unassigned_strands(design):
    print('Checking design for strands without DNA sequences assigned...')
    # Check for and report segments with haven't been assigned DNA sequences
    b_found_unassigned = False
    for strand in design.strands:
        if '?' in strand.dna_sequence:
            print(f'ERROR: strand has unassigned portion: {strand.dna_sequence}, strand {strand}')
            b_found_unassigned = True
    if not b_found_unassigned:
        print('No unassigned segments found.')

# Check for and report mismatches
def check_scadnano_design_for_mismatches(design):
    print('Checking design for DNA sequence mismatches...')
    b_found_mismatches = False
    helices_mismatches = {}
    for helix_idx in design.helices:
        helix = design.helices[helix_idx]
        helix_len = helix.max_offset - helix.min_offset
        forward_sequence = ['x'] * helix_len
        reverse_sequence = ['x'] * helix_len

        for strand in design.strands:
            for domain in strand.domains:
                if (type(domain) is sc.Domain) and (domain.helix == helix_idx):
                    if domain.forward:
                        forward_sequence[domain.start:domain.end] = domain.dna_sequence
                    else:
                        reverse_sequence[domain.start:domain.end] = reversed(domain.dna_sequence)

        helix_mismatches = []
        for i in range(helix_len):
            f = forward_sequence[i]
            r = reverse_sequence[i]
            if (f != 'x') and (r != 'x') and (f != complementary_bases[r]):
                b_found_mismatches = True
                helix_mismatches.append(i)
        helices_mismatches[helix_idx] = helix_mismatches
    if b_found_mismatches:
        print(f'Mismatches found! {helices_mismatches}')
    else:
        print(f'No mismatches found.')

    ## The following implementation seems like it should work by scadnano documentation - but it doesn't find (at least some) mismatches!
    # b_found_mismatches = False
    # base_pairs_dict = design.base_pairs(True)    # Since this allows for mismatches, it should be larger than or equal
    #                                              # to the following dictionary which doesn't
    # base_pairs_dict2 = design.base_pairs(False)
    #
    # for helix_idx in base_pairs_dict:
    #     offset_list = base_pairs_dict[helix_idx]
    #     if helix_idx not in base_pairs_dict2:
    #         print(f'Mismatches found at helix {helix_idx} and offsets {offset_list}')
    #         b_found_mismatches = True
    #     elif offset_list != base_pairs_dict2[helix_idx]:
    #         mismatching_offsets_list = [item for item in offset_list if item not in base_pairs_dict2[helix_idx]]
    #         print(f'Mismatches found at helix {helix_idx} and offsets {mismatching_offsets_list}')
    #         b_found_mismatches = True
    # if not b_found_mismatches:
    #     print(f'No mismatches found.')

def check_scadnano_design_for_single_stranded_segments(design, ignore_list=[]):
    print('Checking design for single-stranded segments...')
    b_single_stranded_found = False

    unpaired_segments = {}
    for helix_idx in design.helices:
        helix = design.helices[helix_idx]
        helix_len = helix.max_offset - helix.min_offset
        forward_sequence = ['x'] * helix_len
        reverse_sequence = ['x'] * helix_len

        for strand in design.strands:
            for domain in strand.domains:
                if (type(domain) is sc.Domain) and (domain.helix == helix_idx):
                    if domain.forward:
                        forward_sequence[domain.start:domain.end] = domain.dna_sequence
                    else:
                        reverse_sequence[domain.start:domain.end] = reversed(domain.dna_sequence)

        b_forward = False
        b_reverse = False
        forward_unpaired = ['-'] * helix_len
        reverse_unpaired = ['-'] * helix_len
        for i in range(helix_len):
            f = forward_sequence[i]
            r = reverse_sequence[i]
            if (f != 'x') and (r == 'x'):
                b_forward = True
                forward_unpaired[i] = f
            if (f == 'x') and (r != 'x'):
                b_reverse = True
                reverse_unpaired[i] = r
        if b_forward:
            b_single_stranded_found = True
            unpaired_segments[f'Helix{helix_idx}_forward'] = forward_unpaired
        if b_reverse:
            b_single_stranded_found = True
            unpaired_segments[f'Helix{helix_idx}_reverse'] = reverse_unpaired
    if b_single_stranded_found:
        print(f'WARNING: Single-stranded segments found.')
        return unpaired_segments
    else:
        print(f'No single-stranded segments found.')
        return None

def create_single_helix_bundle(cooperativity, spacing, scaffold_sequence=None):
    position = sc.Position3D(0, 0, 0)
    helix_length = cooperativity * spacing * 2

    handle_length = 7
    linker = 'TT'
    linker_length = len(linker)
    extension_length = handle_length + linker_length

    new_design = sc.Design()

    helix_idx = 0
    new_helix = sc.Helix(min_offset=0, max_offset=helix_length, idx=helix_idx, position=position, roll=30)
    new_design.add_helix(helix_idx, new_helix)

    scaffold_domains = []
    for i in range(cooperativity*2):
        start = i * spacing
        end = start + spacing

        scaffold_domain = sc.Domain(helix=helix_idx, forward=True, start=start, end=end)
        scaffold_domains.append(scaffold_domain)

        staple_domains = []

        bound_domain = sc.Domain(helix=helix_idx, forward=False, start=start, end=end)
        staple_domains.append(bound_domain)

        #extension = sc.Extension(num_bases=extension_length, display_length=int(extension_length / 3), display_angle=270)
        #staple_domains.append(extension)

        staple = sc.Strand(domains=staple_domains, name=f'handle_staple{i}')
        new_design.add_strand(staple)

    scaffold = sc.Strand(domains=scaffold_domains, name='scaffold', is_scaffold=True, dna_sequence=scaffold_sequence)
    new_design.add_strand(scaffold)
    new_design.assign_dna(strand=scaffold, sequence=scaffold_sequence, assign_complement=True)

    return new_design


####################
# Slat conventions
output_offsets = {N: 0, E: 1, S: 1, W: 0}
input_offsets = {N: 1, E: 0, S: 0, W: 1}
MAX_VAL = 9999999
MIN_VAL = -MAX_VAL

# Glue 'functions'
in1 = 'in1'
in2 = 'in2'
out1 = 'out1'
out2 = 'out2'
out3 = 'out3'
out4 = 'out4'
out5 = 'out5'
out6 = 'out6'
out7 = 'out7'
out8 = 'out8'
intau = 'intau'
outtau = 'outtau'
outtau1 = 'outtau1'
outtau2 = 'outtau2'
outtau3 = 'outtau3'
outtau4 = 'outtau4'
outtau5 = 'outtau5'
outtau6 = 'outtau6'
linktau = 'linktau'  # linktau is not currently fully supported

glue_functions = [in1, in2, out1, out2, out3, out4, out5, out6, out7, out8, intau, outtau, outtau1, outtau2, outtau3, outtau4, outtau5, outtau6, linktau]

glue_markings = {'in1': {N: 'v', E: '<', S: '^', W: '>'},
                 'in2': {N: 'v', E: '<', S: '^', W: '>'},
                 'out1': {N: '^', E: '>', S: 'v', W: '<'},
                 'out2': {N: '^', E: '>', S: 'v', W: '<'},
                 'out3': {N: '^', E: '>', S: 'v', W: '<'},
                 'out4': {N: '^', E: '>', S: 'v', W: '<'},
                 'out5': {N: '^', E: '>', S: 'v', W: '<'},
                 'out6': {N: '^', E: '>', S: 'v', W: '<'},
                 'out7': {N: '^', E: '>', S: 'v', W: '<'},
                 'out8': {N: '^', E: '>', S: 'v', W: '<'},
                 'intau': {N: 'v', E: '<', S: '^', W: '>'},
                 'outtau': {N: '^', E: '>', S: 'v', W: '<'},
                 'outtau1': {N: '^', E: '>', S: 'v', W: '<'},
                 'outtau2': {N: '^', E: '>', S: 'v', W: '<'},
                 'outtau3': {N: '^', E: '>', S: 'v', W: '<'},
                 'outtau4': {N: '^', E: '>', S: 'v', W: '<'},
                 'outtau5': {N: '^', E: '>', S: 'v', W: '<'},
                 'outtau6': {N: '^', E: '>', S: 'v', W: '<'},
                 'linktau': {N: '^', E: '>', S: '^', W: '>'}}

opp_dir = {N: S, E: W, S: N, W: E}
offset = {N: (0, 1), E: (1, 0), S: (0, -1), W: (-1, 0)}

###################################################
# Currently hard-coded parameters for drawing

squareDim = 10
strokeWidth = 1
svgUnits = px
skinnyFactor = 0.7
textStyle = "font-family:Arial"
defaultFillColor = 'grey'
pdfFont = "Courier"  # "Helvetica"
pdfFontBold = "Helvetica-Bold"
# partial_example_color_list = ["black", "black", "green", "red", "blue", "yellow", "pink", "aqua", "lime", "olive", "black"]

# The following is a dictionary of default glue colors which can be extended and/or modified
slat_color_mapping = {'vertical': 'grey', 'horizontal': 'grey', in1: 'lime', in2: 'yellow',
                      out1: 'pink', out2: 'aqua', out3: 'blue', out4: 'blue', out5: 'blue', out6: 'blue', out7: 'blue', out8: 'blue',
                      outtau: 'red', outtau1: 'red', outtau2: 'red', outtau3: 'red', outtau4: 'red', outtau5: 'red', outtau6: 'red',
                      intau: 'red', linktau: 'olive',
                      'crossing_glue_color': 'red', 'linking_glue_color': 'olive'}


####################
# Helper functions for gadget 'tiles'
def isDuplicateTile(tile: object, tile_dict: dict) -> bool:
    for dict_tile in tile_dict:
        if tile.HasMatchingGlues(tile_dict[dict_tile]):
            return True
    return False


def combineTileDicts(dict1: dict, dict2: dict) -> None:
    for tile in dict1:
        if tile in dict2:
            print(f'Tile with name "{tile}" already exists in dict.')
        elif not isDuplicateTile(dict1[tile], dict2):
            dict2[tile] = dict1[tile]


def getTileInputSides(tile: dict) -> list:
    input_sides = []
    for d in [N, E, S, W]:
        if (d in tile) and (not tile[d] is None):
            if tile[d] in [in1, in2, intau]:
                input_sides.append(d)
    return input_sides


def getTileOutputSides(tile: dict) -> list:
    output_sides = []
    for d in [N, E, S, W]:
        if (d in tile) and (not tile[d] is None):
            if tile[d] in [out1, out2, out3, out4, out5, out6, out7, out8, outtau, outtau1, outtau2, outtau3, outtau4, outtau5, outtau6]:
                output_sides.append(d)
    return output_sides


def getTileConnectionSides(tile: dict) -> list:
    connection_sides = []
    for d in [N, E, S, W]:
        if (d in tile) and (not tile[d] is None):
            if ((type(tile[d]) == tuple) and (tile[d][1] in [intau, outtau])):
                connection_sides.append(d)
    return connection_sides


def getTileLinkSides(tile: dict) -> list:
    link_sides = []
    for d in [N, E, S, W]:
        if (d in tile) and (not tile[d] is None):
            if ((d in tile) and (tile[d] == linktau)):
                link_sides.append((d, linktau))
    return link_sides


def areNeighbors(coords1: tuple, coords2: tuple) -> bool:
    if (coords1 is None) or (coords2 is None):
        return False
    x_diff = abs(coords1[0] - coords2[0])
    y_diff = abs(coords1[1] - coords2[1])
    if ((x_diff + y_diff) == 1) and (x_diff == 0 or y_diff == 0):
        return True
    return False


def areConnected(coords1: tuple, coords2: tuple, gadget_tile_list: list) -> bool:
    if not areNeighbors(coords1, coords2):
        return False

    tile1 = None
    tile2 = None
    for tile in gadget_tile_list:
        if tile['coords'] == coords1:
            tile1 = tile
        elif tile['coords'] == coords2:
            tile2 = tile
    if tile1 is None or tile2 is None:
        return False
    x_diff = coords1[0] - coords2[0]
    y_diff = coords1[1] - coords2[1]
    direction_1_to_2 = None
    for d in offset:
        if offset[d] == (x_diff, y_diff):
            direction_1_to_2 = opp_dir[d]
    direction_2_to_1 = opp_dir[direction_1_to_2]
    if (direction_1_to_2 in tile1) and (tile1[direction_1_to_2] is not None) and (
            type(tile1[direction_1_to_2]) == tuple):
        if tile1[direction_1_to_2][1] == intau:
            if (direction_2_to_1 in tile2) and (tile2[direction_2_to_1] is not None) and (
                    type(tile2[direction_2_to_1]) == tuple):
                if tile2[direction_2_to_1][1] == outtau:
                    return True
        elif tile1[direction_1_to_2][1] == outtau:
            if (direction_2_to_1 in tile2) and (tile2[direction_2_to_1] is not None) and (
                    type(tile2[direction_2_to_1]) == tuple):
                if tile2[direction_2_to_1][1] == intau:
                    return True
    return False


def hasConnectionInDirection(coords: tuple, direction: int, gadget_tile_list: list) -> bool:
    nbr_coords = (coords[0] + offset[direction][0], coords[1] + offset[direction][1])
    for tile in gadget_tile_list:
        if tile['coords'] == nbr_coords:
            if areConnected(coords, nbr_coords, gadget_tile_list):
                return True
    return False


def sideIsOutputOrConnection(d: int, tile: dict) -> bool:
    return ((d in tile) and (tile[d] is not None) and (
                tile[d] in [out1, out2, out3, out4, out5, out6, out7, out8, outtau, outtau1, outtau2, outtau3, outtau4, outtau5, outtau6, linktau] or type(tile[d]) == tuple))


def getGadgetTile(coords: tuple, gadget_tile_list: list) -> dict:
    for tile in gadget_tile_list:
        if tile['coords'] == coords:
            return tile
    return None


####################
# Slat classes
class LibSlat():
    def __init__(self, slat_name: str, start_coord: tuple, b_horizontal: bool, scale_factor: int, b_seed_only: bool = False):
        self.name = slat_name
        self.begin = start_coord
        if b_horizontal:
            self.end = (start_coord[0] + scale_factor - 1, start_coord[1])
        else:
            self.end = (start_coord[0], start_coord[1] + scale_factor - 1)
        self.b_horizontal = b_horizontal
        self.color = None
        self.scale_factor = scale_factor
        self.io_locs = []
        self.b_seed_only = b_seed_only

    def __str__(self) -> str:
        out_str = f'{self.name}: {self.begin}->{self.end}'
        for r in self.io_locs:
            out_str = f'{out_str}, [{r[0]} {r[1]}]'
        return out_str

    def equals(self, other, b_ignore_nonfunctional=True):
        if (not b_ignore_nonfunctional) and (self.name != other.name):
            return False
        elif self.begin != other.begin:
            return False
        elif self.end != other.end:
            return False
        elif self.b_horizontal != other.b_horizontal:
            return False
        elif (not b_ignore_nonfunctional) and (self.color != other.color):
            return False
        elif self.scale_factor != other.scale_factor:
            return False
        elif (not b_ignore_nonfunctional) and (self.b_seed_only != other.b_seed_only):
            return False
        elif len(self.io_locs) != len(other.io_locs):
            return False
        else:
            for i in range(len(self.io_locs)):
                if self.io_locs[i] != other.io_locs[i]:
                    return False
        return True

    def copy(self):
        slat_in = LibSlat(self.name, self.begin, self.b_horizontal, self.scale_factor, self.b_seed_only)
        slat_in.end = self.end
        slat_in.color = self.color
        slat_in.io_locs = copy.deepcopy(self.io_locs)
        return slat_in

    def getLength(self) -> int:
        if self.b_horizontal:
            return self.end[0] - self.begin[0] + 1
        else:
            return self.end[1] - self.begin[1] + 1

    def getDims(self) -> tuple:
        return (self.begin, self.end)

    def translate(self, offset_coords: tuple) -> None:
        self.begin = (self.begin[0] + offset_coords[0], self.begin[1] + offset_coords[1])
        self.end = (self.end[0] + offset_coords[0], self.end[1] + offset_coords[1])
        new_io_locs = []
        for io_loc in self.io_locs:
            new_io_locs.append(
                ((io_loc[0][0] + offset_coords[0], io_loc[0][1] + offset_coords[1]), io_loc[1], io_loc[2]))
        self.io_locs = new_io_locs

    def setIOLoc(self, coord: tuple, io_type: int, color: str = None, b_is_crossing: bool = False) -> object:
        if (coord[0] >= self.begin[0]) and (coord[1] >= self.begin[1]) and (coord[0] <= self.end[0]) and (
                coord[1] <= self.end[1]):
            for i in range(len(self.io_locs)):
                if self.io_locs[i][0] == coord:
                    self.io_locs[i] = (coord, io_type, color, b_is_crossing)
                    return
            self.io_locs.append((coord, io_type, color, b_is_crossing))
        else:
            print(f'ERROR: Attempt to set IO in invalid loction of {coord} for Slat {self.begin}->{self.end}')

    def getIOLoc(self, coord: tuple) -> tuple:
        for loc in self.io_locs:
            if loc[0] == coord:
                return loc
        return None

    def getLocByIndex(self, index: int) -> tuple:
        if index > self.getLength():
            return None
        coord = (self.begin[0] + index, self.begin[1]) if self.b_horizontal else (self.begin[0], self.begin[1] + index)
        io = self.getIOLoc(coord)
        if io is None:
            return (coord, None)
        else:
            return (coord, io[1])

    def getIOLocByIndex(self, index: int) -> tuple:
        if index > self.getLength():
            return None
        coord = (self.begin[0] + index, self.begin[1]) if self.b_horizontal else (self.begin[0], self.begin[1] + index)
        io = self.getIOLoc(coord)
        if io is None:
            return None
        else:
            return io

    def setColor(self, color, b_override_io_locs=False):
        self.color = color
        if b_override_io_locs:
            for index in range(self.getLength()):
                io = self.getIOLocByIndex(index)
                if not io is None:
                    self.setIOLoc(io[0], io[1], None, io[3])

    def getDomainLabelSet(self):
        domain_set = set()
        for index in range(self.getLength()):
            io = self.getIOLocByIndex(index)
            if not io is None:
                domain_set.add(io[1])
        return domain_set

    def extend(self, coord: int, io_type: int = None) -> None:
        if (self.b_horizontal and (coord[1] != self.begin[1])) or (
                not self.b_horizontal and (coord[0] != self.begin[0])):
            print(f'ERROR: Invalid extension of Slat, begin: {self.begin}, end: {self.end}, coord: {coord}')
        else:
            if ((coord[0] < self.begin[0]) or (coord[1] < self.begin[1])):
                (curr_x, curr_y) = (coord[0], coord[1])
                (last_x, last_y) = (self.begin[0], self.begin[1])
                if self.b_horizontal:
                    last_x = last_x - 1
                else:
                    last_y = last_y - 1
                self.begin = coord
            elif ((coord[0] > self.end[0]) or (coord[1] > self.end[1])):
                (curr_x, curr_y) = (self.end[0], self.end[1])
                (last_x, last_y) = (coord[0], coord[1])
                if self.b_horizontal:
                    curr_x = curr_x + 1
                else:
                    curr_y = curr_y + 1
                self.end = coord

            if not io_type is None:
                (x_add, y_add) = (1, 0) if self.b_horizontal else (0, 1)
                while (curr_x, curr_y) <= (last_x, last_y):
                    self.setIOLoc((curr_x, curr_y), io_type, None)
                    (curr_x, curr_y) = (curr_x + x_add, curr_y + y_add)

    def writeAsXML(self, xml_parent: object = None, color_mapping: dict = slat_color_mapping) -> object:
        if xml_parent is None:
            xml_parent = ET.Element('PolyominoType')
            elt_ptype = xml_parent
        else:
            elt_ptype = ET.SubElement(xml_parent, 'PolyominoType')

        elt_name = ET.SubElement(elt_ptype, 'name')
        elt_name.text = self.name
        elt_orient = ET.SubElement(elt_ptype, 'b_horizontal')
        elt_orient.text = 'True' if self.b_horizontal else 'False'

        if not self.color is None:
            print(f'slat has color {self.color}')

        # if not self.concentration is None:
        #    elt_concentration = ET.SubElement(elt_ptype, 'concentration')
        #    elt_concentration.text = str(self.concentration)
        # if not self.placement_hint is None:
        #    elt_placement_hint = ET.SubElement(elt_ptype, 'placement_hint')
        #    elt_placement_hint.text = str(self.placement_hint)
        elt_seed = ET.SubElement(elt_ptype, 'b_seed_only')
        elt_seed.text = 'True' if self.b_seed_only else 'False'

        elt_blocks = ET.SubElement(elt_ptype, 'blocks')

        curr_coord = self.begin
        for i in range(self.getLength()):
            loc = self.getLocByIndex(i)
            coord = loc[0]
            domain = loc[1]

            elt_block = ET.SubElement(elt_blocks, 'block')
            elt_coords = ET.SubElement(elt_block, 'coords')
            elt_coords.text = f'{coord}'

            io_loc = self.getIOLoc(coord)
            if not io_loc is None:
                fillColor = self.color #defaultFillColor
                if not io_loc[2] is None:
                    fillColor = io_loc[2]
                elif io_loc[1] in color_mapping:
                    fillColor = color_mapping[io_loc[1]]
                if not fillColor is None:
                    elt_color = ET.SubElement(elt_block, 'color')
                    elt_color.text = f'{fillColor}'

            if not domain is None:
                elt_domains = ET.SubElement(elt_block, 'domains')
                elt_domain = ET.SubElement(elt_domains, 'domain')
                elt_label = ET.SubElement(elt_domain, 'label')
                elt_label.text = f'{domain}'

                elt_direction = ET.SubElement(elt_domain, 'direction')
                elt_direction.text = 'D' if self.b_horizontal else 'U'  # Can flip this to have vertical slats on bottom

                elt_strength = ET.SubElement(elt_domain, 'strength')
                elt_strength.text = str(1)
        return xml_parent


class SlatGroup:
    def __init__(self, slats=None):
        if slats is None:
            self.slats = []
        else:
            self.slats = slats

    def getDomainLabelSet(self):
        domain_set = set()
        for slat in self.slats:
            slat_doamin_set = slat.getDomainLabelSet()
            domain_set = domain_set.union(slat_domain_set)
        return domain_set

class GadgetSlatGroup:
    def __init__(self, gadget_name: str = None, horizontal_slats: SlatGroup = None, vertical_slats: SlatGroup = None, gadget_tile_list: list = None) -> None:
        self.gadget_name = gadget_name
        if horizontal_slats is None:
            horizontal_slats = SlatGroup()
        if vertical_slats is None:
            vertical_slats = SlatGroup()
        self.horizontal_slat_group = horizontal_slats
        self.vertical_slat_group = vertical_slats
        self.gadget_tile_list = gadget_tile_list

    def getDims(self) -> tuple:
        (minX, minY) = (MAX_VAL, MAX_VAL)
        (maxX, maxY) = (MIN_VAL, MIN_VAL)

        for slats in [self.horizontal_slat_group.slats, self.vertical_slat_group.slats]:
            for slat in slats:
                if slat.begin[0] < minX:
                    minX = slat.begin[0]
                if slat.begin[1] < minY:
                    minY = slat.begin[1]
                if slat.end[0] > maxX:
                    maxX = slat.end[0]
                if slat.end[1] > maxY:
                    maxY = slat.end[1]

        return ((minX, minY), (maxX, maxY))

    def translate(self, offset_coords: tuple) -> None:
        for slats in [self.horizontal_slat_group.slats, self.vertical_slat_group.slats]:
            for slat in slats:
                slat.translate(offset_coords)

    def getMacrotileLocationCoords(self, coord: tuple) -> tuple:
        scale_factor = None
        if len(self.horizontal_slat_group.slats) > 0:
            scale_factor = self.horizontal_slat_group.slats[0].scale_factor
        elif len(self.vertical_slat_group.slats) > 0:
            scale_factor = self.vertical_slat_group.slats[0].scale_factor
        if scale_factor is None:
            print(f'ERROR: No slats in GadgetSlatGroup for {self.gadget_name} in getMacrotileLocationIndex')
            return None
        x_coord = coord[0]
        y_coord = coord[1]
        mod_coord = (math.floor(x_coord / scale_factor), math.floor(y_coord / scale_factor))
        return mod_coord

    def getMacrotileLocationIndex(self, coord: tuple) -> int:
        scale_factor = None
        if len(self.horizontal_slat_group.slats) > 0:
            scale_factor = self.horizontal_slat_group.slats[0].scale_factor
        elif len(self.vertical_slat_group.slats) > 0:
            scale_factor = self.vertical_slat_group.slats[0].scale_factor
        if scale_factor is None:
            print(f'ERROR: No slats in GadgetSlatGroup for {self.gadget_name} in getMacrotileLocationIndex')
            return None
        mod_coord = (coord[0] % scale_factor, coord[1] % scale_factor)
        mt_loc_index = (mod_coord[1] * scale_factor) + mod_coord[0]
        return mt_loc_index

    def applyGlueDict(self, glue_dict: dict, motif: list = None) -> None:
        for glue_type in glue_dict:
            glue_label = glue_dict[glue_type]
            for slats in [self.horizontal_slat_group.slats, self.vertical_slat_group.slats]:
                glue_locations = []
                for slat_num in range(len(slats)):
                    for io_loc in slats[slat_num].io_locs:
                        if io_loc[1] == glue_type:
                            glue_locations.append((slat_num, io_loc[0]))
                for glue_location in glue_locations:
                    if motif is None:
                        loc_index = self.getMacrotileLocationIndex(glue_location[1])
                    else:
                        motif_letter1 = motif[glue_location[1][0] % len(motif)]
                        motif_letter2 = motif[glue_location[1][1]%len(motif)]
                        #print(f'glue_location: {glue_location}, motif_letter: {(motif_letter1,motif_letter2)}')
                        loc_index = f'{motif_letter1}{motif_letter2}'
                    color = slat_color_mapping[glue_type]
                    # Mark output versions of glues as complements
                    comp = ''
                    if slats == self.horizontal_slat_group.slats:
                        comp = '*'
                    # if glue_type in [out1, out2, out3, out4, out5, out6, out7, out8, outtau, outtau1, outtau2 outtau3, outtau4, outtau5 outtau6]:
                    #     comp = '*'
                    # elif (slats == self.vertical_slat_group.slats) and glue_type in [linktau]:
                    #     comp = '*'
                    slats[glue_location[0]].setIOLoc(glue_location[1], f'{glue_label}_{loc_index}{comp}', color)

    def applyColor(self, color, b_override_all=False):
        for hor_slat in self.horizontal_slat_group.slats:
            hor_slat.setColor(color, b_override_all)
        for vert_slat in self.vertical_slat_group.slats:
            vert_slat.setColor(color, b_override_all)
    def addLinkingGlues(self, linking_locs: list, motif: list = None) -> None:
        # Assign glues to slat positions at each linking location
        for hor_slat in self.horizontal_slat_group.slats:
            for i in range(hor_slat.getLength()):
                loc = hor_slat.getLocByIndex(i)[0]
                loc_coords = self.getMacrotileLocationCoords(loc)
                for linking_loc in linking_locs:
                    if linking_loc[0] == loc_coords:
                        #print(f'horizontal linker at {loc} {linking_loc[1][0][1]}')
                        hor_slat.setIOLoc((loc), linking_loc[1][0][1], slat_color_mapping['linking_glue_color']) # ASSUMING ONLY A SINGLE LINKING DIRECTION HERE!!!
        for vert_slat in self.vertical_slat_group.slats:
            for i in range(vert_slat.getLength()):
                loc = vert_slat.getLocByIndex(i)[0]
                loc_coords = self.getMacrotileLocationCoords(loc)
                for linking_loc in linking_locs:
                    if linking_loc[0] == loc_coords:
                        #print(f'vertical linker at {loc} {linking_loc[1][0][1]}')
                        vert_slat.setIOLoc((loc), linking_loc[1][0][1], slat_color_mapping['linking_glue_color']) # ASSUMING ONLY A SINGLE LINKING DIRECTION HERE!!!


    def addCrossingGlues(self, motif: list = None) -> None:
        # First, find the crossing locations
        slat_coords = dict()
        crossing_count = 1
        for hor_slat in self.horizontal_slat_group.slats:
            for i in range(hor_slat.getLength()):
                loc = hor_slat.getLocByIndex(i)[0]
                if loc in slat_coords:
                    print(f'ERROR: coords {loc} found twice in horizontal slats in addCrossingGlues.')
                else:
                    slat_coords[loc] = 0
        for vert_slat in self.vertical_slat_group.slats:
            for i in range(vert_slat.getLength()):
                loc = vert_slat.getLocByIndex(i)[0]
                if loc in slat_coords:
                    if slat_coords[loc] != 0:
                        print(f'ERROR: coords {loc} more than twice in slat groups in addCrossingGlues.')
                    else:
                        slat_coords[loc] = crossing_count
                        crossing_count = crossing_count + 1

        # Now assign glues to slat positions at each crossing location
        for hor_slat in self.horizontal_slat_group.slats:
            for i in range(hor_slat.getLength()):
                loc = hor_slat.getLocByIndex(i)[0]
                if (loc in slat_coords) and (slat_coords[loc] != 0):
                    if not motif is None:
                        motif_letter1 = motif[loc[0] % len(motif)]
                        motif_letter2 = motif[loc[1] % len(motif)]
                        #print(f'loc: {loc}, motif_letter: {(motif_letter1,motif_letter2)}')
                        macrotile_loc = self.getMacrotileLocationCoords(loc)
                        loc_index = f'{motif_letter1}{motif_letter2}'
                        hor_slat.setIOLoc((loc), f'{self.gadget_name}x{macrotile_loc}{loc_index}*'.replace(' ', ''),
                                          slat_color_mapping['crossing_glue_color'], True)
                    else:
                        loc_index = self.getMacrotileLocationIndex(loc)
                        hor_slat.setIOLoc((loc), f'{self.gadget_name}x{loc}{loc_index}*',
                                          slat_color_mapping['crossing_glue_color'], True)
        for vert_slat in self.vertical_slat_group.slats:
            for i in range(vert_slat.getLength()):
                loc = vert_slat.getLocByIndex(i)[0]
                if (loc in slat_coords) and (slat_coords[loc] != 0):
                    if not motif is None:
                        motif_letter1 = motif[loc[0] % len(motif)]
                        motif_letter2 = motif[loc[1]%len(motif)]
                        #print(f'loc: {loc}, motif_letter: {(motif_letter1,motif_letter2)}')
                        macrotile_loc = self.getMacrotileLocationCoords(loc)
                        loc_index = f'{motif_letter1}{motif_letter2}'
                        vert_slat.setIOLoc((loc), f'{self.gadget_name}x{macrotile_loc}{loc_index}'.replace(' ', ''),
                                          slat_color_mapping['crossing_glue_color'], True)
                    else:
                        loc_index = self.getMacrotileLocationIndex(loc)
                        vert_slat.setIOLoc((loc), f'{self.gadget_name}x{loc}{loc_index}',
                                           slat_color_mapping['crossing_glue_color'], True)


def ExtendSlats(horizontal_slat_dict: dict, vertical_slat_dict: dict, new_length: int) -> (dict, dict):
    slat_length_set = set()
    for slats in [horizontal_slat_dict, vertical_slat_dict]:
        for slat_name in slats:
            slat = slats[slat_name]
            slat_length = slat.getLength()
            slat_length_set.add(slat_length)

    slat_lengths = sorted(list(slat_length_set))
    shortest_length = slat_lengths[0]
    if (shortest_length > new_length):
        print(f'ERROR: In GadgetSlatGroup::extendSlats, the requested length of {new_length} is greater than the current shortest length of {shortest_length}. Skipping extension.')
        return (horizontal_slat_dict, vertical_slat_dict)

    handle_location_ratio = math.ceil(new_length / shortest_length)

    new_horizontal_slat_dict = dict()
    new_vertical_slat_dict = dict()

    for slat_name in horizontal_slat_dict:
        slat = horizontal_slat_dict[slat_name]
        new_slat_length = slat.getLength() * handle_location_ratio
        b_do_converion = False
        orig_slat = None
        if type(slat) == LibSlat:
            orig_slat = slat
            slat = GetTASSlatTypeFromLibSlatType(slat, slat.b_seed_only, b_do_initialize=False)
            b_do_converion = True

        new_slat = slat.Copy()
        new_slat.blocks = []
        new_slat.domains = dict()
        new_slat.colors = dict()
        min_x = 9999999
        min_y = 9999999
        min_z = 9999999
        for block in slat.blocks:
            if block[0] < min_x:
                min_x = block[0]
            if block[1] < min_y:
                min_y = block[1]
            if block[2] < min_z:
                min_z = block[2]

        for i in range(new_slat_length):
            new_slat.blocks.append(((min_x * handle_location_ratio) + i, min_y * handle_location_ratio, min_z))
        for i in range(len(slat.blocks)):
            if slat.blocks[i] in slat.domains:
                new_slat.domains[((min_x * handle_location_ratio) + (i * handle_location_ratio), min_y * handle_location_ratio, min_z)] = slat.domains[slat.blocks[i]]
            if slat.blocks[i] in slat.colors:
                new_slat.colors[((min_x * handle_location_ratio) + (i * handle_location_ratio), min_y * handle_location_ratio, min_z)] = slat.colors[slat.blocks[i]]
        new_slat.ensure_3D()
        foo = new_slat.getBoundingBox()
        new_slat.ComputeReferencePoint()
        new_slat.ComputePerimeterOffsets(new_slat.dim_restrictions)

        if b_do_converion:
            new_slat = GetLibSlatTypeFromTASSlatType(new_slat)
        new_horizontal_slat_dict[slat_name] = new_slat

    for slat_name in vertical_slat_dict:
        slat = vertical_slat_dict[slat_name]
        new_slat_length = slat.getLength() * handle_location_ratio
        b_do_converion = False
        if type(slat) == LibSlat:
            slat = GetTASSlatTypeFromLibSlatType(slat, slat.b_seed_only, b_do_initialize=False)
            b_do_converion = True

        new_slat = slat.Copy()
        new_slat.blocks = []
        new_slat.domains = dict()
        new_slat.colors = dict()
        min_x = 9999999
        min_y = 9999999
        min_z = 9999999
        for block in slat.blocks:
            if block[0] < min_x:
                min_x = block[0]
            if block[1] < min_y:
                min_y = block[1]
            if block[2] < min_z:
                min_z = block[2]

        for i in range(new_slat_length):
            new_slat.blocks.append((min_x * handle_location_ratio, min_y + i, min_z))
        for i in range(len(slat.blocks)):
            if slat.blocks[i] in slat.domains:
                new_slat.domains[(min_x * handle_location_ratio, min_y + (i * handle_location_ratio), min_z)] = slat.domains[slat.blocks[i]]
            if slat.blocks[i] in slat.colors:
                new_slat.colors[(min_x * handle_location_ratio, min_y + (i * handle_location_ratio), min_z)] = slat.colors[slat.blocks[i]]
        new_slat.ensure_3D()
        foo = new_slat.getBoundingBox()
        new_slat.ComputeReferencePoint()
        new_slat.ComputePerimeterOffsets(new_slat.dim_restrictions)

        if b_do_converion:
            new_slat = GetLibSlatTypeFromTASSlatType(new_slat)
        new_vertical_slat_dict[slat_name] = new_slat

    return (new_horizontal_slat_dict, new_vertical_slat_dict)


def ExtendGadgetSlats(gadget_slat_groups, new_length):
    for gadget_slat_group in gadget_slat_groups:
        horizontal_slat_dict = dict()
        vertical_slat_dict = dict()
        for slats in [gadget_slat_group.vertical_slat_group.slats, gadget_slat_group.horizontal_slat_group.slats]:
            for slat in slats:
                if slat.b_horizontal:
                    horizontal_slat_dict[slat.name] = slat
                else:
                    vertical_slat_dict[slat.name] = slat
        (new_horizontal_slat_dict, new_vertical_slat_dict) = ExtendSlats(horizontal_slat_dict, vertical_slat_dict, new_length)

        gadget_slat_group.horizontal_slat_group.slats = list(new_horizontal_slat_dict.values())
        gadget_slat_group.vertical_slat_group.slats = list(new_vertical_slat_dict.values())



# MJP TODO: GetSlatGroup needs to be modified to account for slat type reuse and ordering
def GetSlatGroup(gadget_name: str, slat_count: int, prelim_slat: list, b_horizontal: bool, gadget_tile_list: list, scale_factor: int, motif: list = None) -> SlatGroup:
    prelim_slat = sorted(prelim_slat)
    in_coord = None
    out_coords = []
    all_coords = list(prelim_slat)
    slats = []
    dirs = [E, W] if b_horizontal else [N, S]

    motif_set1 = set()
    motif_set2 = set()
    motif_sets = [motif_set1, motif_set2]
    motif_inputs = {N: motif_set1, S: motif_set2, E: motif_set1, W: motif_set2}
    motif_outputs = {S: motif_set1, N: motif_set2, W: motif_set1, E: motif_set2}
    curr_set = 0
    if not motif is None:
        motif_set = set(motif)
        motif_set = sorted(motif_set)
        if (len(motif_set) % 2) != 0:
            print(f'ERROR:In GetSlatGroup, there must be an even number of types in the motif!')
        else:
            for m in motif_set:
                motif_sets[curr_set].add(m)
                curr_set = (curr_set + 1) % 2
        #print(f'motif sets: {motif_sets}')

    #print(f'[DEBUG] gadget_name: {gadget_name} prelim_slat = {prelim_slat}, slat_count: {slat_count}, motif: {motif}')
    for coords in prelim_slat:
        tile = getGadgetTile(coords, gadget_tile_list)

        bound = scale_factor if motif is None else len(motif)

        for i in range(bound):
            if b_horizontal:
                x_add = 0
                y_add = i
            else:
                x_add = i
                y_add = 0
            slat_x = coords[0] * scale_factor + x_add
            slat_y = coords[1] * scale_factor + y_add
            slat_coords = (slat_x, slat_y)
            if len(slats) < (i + 1):
                orientation = 'h' if b_horizontal else 'v'

                name_end = slat_count if motif is None else motif[i]

                slat_name = f'{gadget_name}{orientation}{name_end}'
                slat_count = slat_count + 1
                new_slat = LibSlat(slat_name, slat_coords, b_horizontal, scale_factor)
                slats.append(new_slat)
            else:
                if b_horizontal:
                    if slat_x >= 0:
                        slat_x = slat_x + scale_factor - 1
                else:
                    if slat_y >= 0:
                        slat_y = slat_y + scale_factor - 1
                slats[i].extend((slat_x, slat_y))

        input_sides = getTileInputSides(tile)
        if len(input_sides) > 0:
            if not in_coord is None:
                print(f'ERROR: slat already has an input location!')
            in_coord = coords

            for in_side in input_sides:
                for i in range(scale_factor):
                    b_gets_input = False
                    if (tile[in_side] == 'intau'):
                        b_gets_input = True
                    elif not motif is None:
                        motif_letter = motif[i]
                        if motif_letter in motif_inputs[in_side]:
                            b_gets_input = True
                            #print(f' get input from side {in_side} with motif letter {motif_letter}')
                    elif (((i + input_offsets[in_side]) % 2) == 0):
                        b_gets_input = True

                    if b_gets_input:
                        x_add = i if b_horizontal else 0
                        y_add = 0 if b_horizontal else i
                        slat_x = in_coord[0] * scale_factor + x_add
                        slat_y = in_coord[1] * scale_factor + y_add

                        if b_horizontal:
                            # TODO: may need an adjustment if input is located in a negative coordinate
                            # if slat_y >= 0:
                            #    slat_y = slat_y + scale_factor - 1
                            for i in range(scale_factor):
                                slats[i].setIOLoc((slat_x, slat_y), tile[in_side])
                                slat_y = slat_y + 1
                        else:
                            # if slat_x >= 0:
                            #    slat_x = slat_x + scale_factor - 1
                            for i in range(scale_factor):
                                slats[i].setIOLoc((slat_x, slat_y), tile[in_side])
                                slat_x = slat_x + 1

        output_sides = getTileOutputSides(tile)
        for os in output_sides:
            if os in dirs:
                out_coord = (coords[0] + offset[os][0], coords[1] + offset[os][1])
                out_coords.append(out_coord)

                for i in range(scale_factor):
                    b_gives_output = False
                    if (tile[os] in ['outtau', 'outtau1', 'outtau2', 'outtau3', 'outtau4', 'outtau5', 'outtau6']):
                        b_gives_output = True
                    elif not motif is None:
                        motif_letter = motif[i]
                        if motif_letter in motif_outputs[os]:
                            b_gives_output = True
                    elif (((i + output_offsets[os]) % 2) == 0):
                        b_gives_output = True

                    if b_gives_output:
                        x_add = 0 if b_horizontal else i
                        y_add = i if b_horizontal else 0
                        slat_x = out_coord[0] * scale_factor + x_add
                        slat_y = out_coord[1] * scale_factor + y_add

                        # Ensure that extra addition happens if the extension is in the positive direction
                        if b_horizontal:
                            if slat_x > slats[i].end[0]:
                                slat_x = slat_x + scale_factor - 1
                        else:
                            if slat_y > slats[i].end[1]:
                                slat_y = slat_y + scale_factor - 1
                        # print(f'  for output coords {out_coords}, slat_x = {slat_x}')
                        # slat_coords = (out_coord[0]*scale_factor+x_add, out_coord[1]*scale_factor+y_add)
                        slats[i].extend((slat_x, slat_y), tile[os])
    all_coords = all_coords + out_coords
    # print(f'prelim slat {prelim_slat} has input coord {in_coord}, output coords {out_coords}, and all coords {all_coords}')
    slat_group = SlatGroup()
    for s in slats:
        slat_length = s.getLength()
        if slat_length > scale_factor:  # get rid of slats of length == scale_factor
            slat_group.slats.append(s)
        # else:
        #    print(f'Note: Excluding slat of length {slat_length}: {s}')

    return slat_group


def GetGadgetSlats(gadget_tile_list: list, scale_factor: int, gadget_name: str, glue_dict: dict, motif: list = None) -> GadgetSlatGroup:
    horiz_slats = {}
    vert_slats = {}
    #print(f'DEBUG: In GetGadgetSlats for {gadget_name} with motif {motif}')

    linking_locs = []
    for tile in gadget_tile_list:
        coords = tile['coords']
        links = getTileLinkSides(tile)
        if len(links) > 0:
            linking_locs.append((coords, links))

        # Find horizontal slats
        if sideIsOutputOrConnection(E, tile) or sideIsOutputOrConnection(W, tile):
            if not coords[1] in horiz_slats:
                horiz_slats[coords[1]] = [tile]
            else:
                horiz_slats[coords[1]].append(tile)

        # Find vertical slats
        if sideIsOutputOrConnection(N, tile) or sideIsOutputOrConnection(S, tile):
            if not coords[0] in vert_slats:
                vert_slats[coords[0]] = [tile]
            else:
                vert_slats[coords[0]].append(tile)

    for coord in horiz_slats:
        horiz_slats[coord] = sorted(horiz_slats[coord], key=lambda tile: tile['coords'][0])
    for coord in vert_slats:
        vert_slats[coord] = sorted(vert_slats[coord], key=lambda tile: tile['coords'][1])

    #print(f'DEBUG: {len(horiz_slats)} horiz_slats: {horiz_slats}\n')
    # print(f'DEBUG: {len(vert_slats)} vert_slats: {vert_slats}\n')

    hs = []
    for y_coord in horiz_slats:
        curr_coords = horiz_slats[y_coord][0]['coords']
        slat_coords = [curr_coords]
        for nc in horiz_slats[y_coord][1:]:
            next_coords = nc['coords']
            if areConnected(curr_coords, next_coords, gadget_tile_list):
                curr_coords = next_coords
                slat_coords.append(curr_coords)
            else:
                hs.append(slat_coords)
                slat_coords = [next_coords]
                curr_coords = next_coords
        hs.append(slat_coords)
    #print(f'hs: {hs}')

    vs = []
    for x_coord in vert_slats:
        curr_coords = vert_slats[x_coord][0]['coords']
        slat_coords = [curr_coords]
        for nc in vert_slats[x_coord][1:]:
            next_coords = nc['coords']
            if areConnected(curr_coords, next_coords, gadget_tile_list):
                curr_coords = next_coords
                slat_coords.append(curr_coords)
            else:
                vs.append(slat_coords)
                slat_coords = [next_coords]
                curr_coords = next_coords
        vs.append(slat_coords)

    horiz_slat_group = SlatGroup()
    slat_count = 1
    for h in hs:
        gadget_name_with_coord = f'{gadget_name}{h}'
        h_group = GetSlatGroup(gadget_name_with_coord, slat_count, h, True, gadget_tile_list, scale_factor, motif)
        horiz_slat_group.slats = horiz_slat_group.slats + h_group.slats
        slat_count = slat_count + len(horiz_slat_group.slats)

    vert_slat_group = SlatGroup()
    for v in vs:
        gadget_name_with_coord = f'{gadget_name}{v}'
        v_group = GetSlatGroup(gadget_name_with_coord, slat_count, v, False, gadget_tile_list, scale_factor, motif)
        vert_slat_group.slats = vert_slat_group.slats + v_group.slats
        slat_count = slat_count + len(vert_slat_group.slats)

    gadget_slats = GadgetSlatGroup(gadget_name, horiz_slat_group, vert_slat_group, gadget_tile_list)
    if len(linking_locs) > 0:
        gadget_slats.addLinkingGlues(linking_locs, motif)
    gadget_slats.applyGlueDict(glue_dict, motif)
    gadget_slats.addCrossingGlues(motif)
    if 'color' in glue_dict:
        gadget_slats.applyColor(glue_dict['color'], b_override_all=True)

    return gadget_slats


def GetGadgetTiles(gadget_name: str, gadget_tile_list: list, tile_label: str, tile_color: str, glue_dict: dict) -> dict:
    tile_dict = dict()

    tile_count = 1
    for tile in gadget_tile_list:
        inputs = '' if (in1 not in glue_dict or glue_dict[in1] is None) else str(glue_dict[in1])
        inputs = inputs + ('' if (in2 not in glue_dict or glue_dict[in2] is None) else str(glue_dict[in2]))
        tile_name = f'{gadget_name}{inputs}{tile_count}'
        if tile_name in tile_dict:
            print(f'Tile with name "{tile_name}" already exists')
        else:
            t = PyTASTileType()
            t.name = tile_name
            t.label = tile_label
            for direction in [N, E, S, W]:
                side = tile[direction]
                if side in glue_functions:
                    if (side is not None) and (side in glue_dict) and (glue_dict[side] is not None):
                        t.sides[direction].label = f'{glue_dict[side]}{glue_markings[side][direction]}'
                        t.sides[direction].strength = 2 if (side in [intau, outtau, outtau1, outtau2, outtau3, outtau4, outtau5, outtau6, linktau]) else 1
                elif type(side) == tuple:
                    t.sides[direction].label = f'{gadget_name}-{side[0]}{glue_markings[side[1]][direction]}'
                    t.sides[direction].strength = 2

            if type(tile_color) == list:
                t.color = f'{tile_color[tile_count - 1]}'
            else:
                t.color = f'{tile_color}'

            if isDuplicateTile(t, tile_dict):
                print(f'Equivalent tile already exists for new tile named "{tile_name}".')
            else:
                tile_dict[tile_name] = t

        tile_count = tile_count + 1

    return tile_dict


def GetMotifLetterColor(motif_letter, motif):
    if motif_letter in motif:
        index = motif.index(motif_letter)
        return color_list[index % len(color_list)]
    return None

def SetSlatTypeToMotifColor(tas_slat_type, lib_slat_type, motif):
    if not tas_slat_type is None:
        motif_letter = tas_slat_type.name[-1]
        color = GetMotifLetterColor(motif_letter, motif)
        if not color is None:
            tas_slat_type.colors = dict()
            tas_slat_type.color = color

    if not lib_slat_type is None:
        motif_letter = lib_slat_type.name[-1]
        color = GetMotifLetterColor(motif_letter, motif)
        if not color is None:
            lib_slat_type.color = color
            for i in range(len(lib_slat_type.io_locs)):
                lib_slat_type.io_locs[i] = (lib_slat_type.io_locs[i][0], lib_slat_type.io_locs[i][1], color, lib_slat_type.io_locs[i][3])

def GetLibSlatTypeFromTASSlatType(tas_slat_type):
    b_horizontal = tas_slat_type.IsHorizontal()
    lib_slat_type = LibSlat(tas_slat_type.name, tas_slat_type.blocks[0], b_horizontal, int(len(tas_slat_type.blocks) / 2), tas_slat_type.b_seed_only)
    lib_slat_type.color = tas_slat_type.color

    start_coords = tas_slat_type.blocks[0]
    count = 0
    for block_coords in tas_slat_type.blocks:
        if b_horizontal:
            lib_slat_type.end = (start_coords[0] + count, start_coords[1])
        else:
            lib_slat_type.end = (start_coords[0], start_coords[1] + count)
        count += 1
        color = None
        if block_coords in tas_slat_type.colors:
            color = tas_slat_type.colors[block_coords]
        if block_coords in tas_slat_type.domains:
            domain = tas_slat_type.domains[block_coords][0]
            lib_slat_type.setIOLoc((block_coords[0], block_coords[1]), domain[0], color, b_is_crossing = False)
    return lib_slat_type

def GetLibSlatTypeFromDomainList(slat_name, start_coord, b_horizontal, domain_list):

    length = len(domain_list)
    lib_slat_type = LibSlat(slat_name=slat_name, start_coord=start_coord, b_horizontal=b_horizontal, scale_factor=length)
    for count in range(length):
        if b_horizontal:
            lib_slat_type.setIOLoc(coord=(start_coord[0]+count, start_coord[1]), io_type=domain_list[count])
        else:
            lib_slat_type.setIOLoc(coord=(start_coord[0], start_coord[1]+count), io_type=domain_list[count])
        count += 1
    return lib_slat_type


def GetTASSlatTypeFromLibSlatType(lib_slat_type, b_seed_only=False, b_do_initialize=True):
    blocks = []
    domains = {}
    colors = {}
    for i in range(lib_slat_type.getLength()):
        curr_coords = (lib_slat_type.begin[0]+i, lib_slat_type.begin[1]) if lib_slat_type.b_horizontal else (lib_slat_type.begin[0], lib_slat_type.begin[1]+i)
        blocks.append(curr_coords)
        curr_loc =  lib_slat_type.getIOLoc((curr_coords[0], curr_coords[1]))
        if not curr_loc is None:
            domain_label = curr_loc[1]
            domain_dir = 'D' if lib_slat_type.b_horizontal else 'U'
            domains[curr_coords] = [(domain_label, domain_dir, 1)]
            color = curr_loc[2]
            if not color is None:
                 #fillColor = color
                colors[curr_coords] = color
            #elif domain_label in color_mapping:
            #     fillColor = color_mapping[io_loc[1]]
    tas_slat_type = TASSlatType(lib_slat_type.name, blocks=blocks, colors=colors, domains=domains)
    tas_slat_type.color = lib_slat_type.color
    if b_do_initialize:
        tas_slat_type.Initialize()
    if b_seed_only:
        tas_slat_type.b_seed_only = True

    #print(f'slat_type {tas_slat_type.name} has blocks {tas_slat_type.blocks}')
    return tas_slat_type

def GetDomainSetFromSlatSystem(slat_system, b_include_seed_types=False):
    domain_set = set()
    for slat_type_name in slat_system.PolyominoType_dict:
        slat_type = slat_system.PolyominoType_dict[slat_type_name]

        if (not b_include_seed_types) or (slat_type.b_seed_only == False):
            domain_label_list = slat_type.get_domain_label_list()
            for domain_label in domain_label_list:
                # Not concerned with both domains of a complementary pair
                if domain_label[-1] != '*':
                    domain_set.add(domain_label)
    return domain_set

###################################################
# Output to files

def OutputSlatsToXML(xml_file_name, gadget_slat_groups, scale_factor, seed_assembly=None, color_mapping=slat_color_mapping):
    # Make copies of slats since they will be translated
    gadget_slat_groups = copy.deepcopy(gadget_slat_groups)
    if not seed_assembly is None:
        seed_assembly = copy.deepcopy(seed_assembly)

    xml_root = ET.Element('PolyominoSystem')
    elt_bt = ET.SubElement(xml_root, 'bindingThreshold')
    elt_bt.text = str(scale_factor)
    elt_rotation = ET.SubElement(xml_root, 'rotation')
    elt_rotation.text = str('False')
    elt_restrictions = ET.SubElement(xml_root, 'dim_restrictions')
    elt_restrictions.text = str('(None,None,[0,2])')
    ##        if not self.Gmc is None:
    ##            elt_gmc = ET.SubElement(xml_root, 'Gmc')
    ##            elt_gmc.text = str(self.Gmc)
    ##        if not self.Gse is None:
    ##            elt_gse = ET.SubElement(xml_root, 'Gse')
    ##            elt_gse.text = str(self.Gse)

    elt_pct_set = ET.SubElement(xml_root, 'PolyominoTypes')

    if type(gadget_slat_groups) != list:
        gadget_slat_groups = [gadget_slat_groups]

    for gadget_slat_group in gadget_slat_groups:
        for slats in [gadget_slat_group.vertical_slat_group.slats, gadget_slat_group.horizontal_slat_group.slats]:
            for slat in slats:
                ((minX, minY), (maxX, maxY)) = slat.getDims()
                slat.translate((-minX, -minY))
                slat.writeAsXML(elt_pct_set, color_mapping)

    if not seed_assembly is None:
        elt_seed = ET.SubElement(xml_root, 'seed')
        elt_polyominoes = ET.SubElement(elt_seed, 'Polyominoes')
        for seed_slat_group in seed_assembly:
            for slats in [seed_slat_group.vertical_slat_group.slats, seed_slat_group.horizontal_slat_group.slats]:
                for slat in slats:
                    elt_polyomino = ET.SubElement(elt_polyominoes, 'Polyomino')

                    elt_ptype_name = ET.SubElement(elt_polyomino, 'PolyominoType')
                    elt_ptype_name.text = slat.name

                    ((minX, minY), (maxX, maxY)) = slat.getDims()
                    elt_translation = ET.SubElement(elt_polyomino, 'translation')
                    zTrans = 0 if slats == seed_slat_group.vertical_slat_group.slats else 1
                    elt_translation.text = f'{(minX, minY, zTrans)}'

    ##    if not self.assembly is None:
    ##        elt_assembly = ET.SubElement(xml_root, 'assembly')
    ##        elt_polyominoes = ET.SubElement(elt_assembly, 'Polyominoes')
    ##        polyominoes = self.assembly.GetPolyominoes()
    ##        for polyomino in polyominoes:
    ##            polyomino.toXML(elt_polyominoes)

    xml_string = ET.tostring(xml_root)
    xml_dom = xml.dom.minidom.parseString(xml_string)

    with open(xml_file_name, "w") as f:
        f.write(xml_dom.toprettyxml())
        print(f"Slats saved to '{xml_file_name}'.")


def DrawSlatsToFiles(gadget_slat_groups, color_mapping=slat_color_mapping, svg_file_name=None,
                     pdf_file_name=None, b_translate_for_grid=True, b_show_labels=True, tile_set=None):
    dwg = None
    c = None
    location_map = {}

    if type(gadget_slat_groups) != list:
        gadget_slat_groups = [gadget_slat_groups]

    if b_translate_for_grid == True:
        max_x_dim = 0
        max_y_dim = 0
        group_num = 0
        for gadget_slat_group in gadget_slat_groups:
            ((minX, minY), (maxX, maxY)) = gadget_slat_group.getDims()
            # Automatically translating to positive coordinates
            gadget_slat_group.translate((-minX, -minY))
            x_dim = maxX - minX
            y_dim = maxY - minY
            if x_dim > max_x_dim:
                max_x_dim = x_dim
            if y_dim > max_y_dim:
                max_y_dim = y_dim
            group_num += 1

        max_x_dim = max_x_dim + 4
        max_y_dim = max_y_dim + 4
        num_cols = math.floor(math.sqrt(group_num))
        num_rows = math.ceil(group_num / num_cols)
        #pageWidth = num_tiles_per_row * CELLSIZE
        #pageHeight = num_rows * CELLSIZE

        #num_rows = int(len(gadget_slat_groups) / row_width) + 1
        #num_cols = min(len(gadget_slat_groups), (len(gadget_slat_groups) % row_width) + 1)
        pageWidth = max_x_dim * num_cols * squareDim
        pageHeight = max_y_dim * num_rows * squareDim

        gadget_group_num = 0
        for gadget_slat_group in gadget_slat_groups:
            row_num = int(gadget_group_num / num_cols)
            col_num = gadget_group_num % num_cols
            gadget_group_num = gadget_group_num + 1

            ((minX, minY), (maxX, maxY)) = gadget_slat_group.getDims()
            gadget_slat_group.translate((col_num * max_x_dim, row_num * max_y_dim))
    else:
        min_x = MAX_VAL
        min_y = MAX_VAL
        max_x = MIN_VAL
        max_y = MIN_VAL
        for gadget_slat_group in gadget_slat_groups:
            ((minX, minY), (maxX, maxY)) = gadget_slat_group.getDims()
            if minX < min_x:
                min_x = minX
            if minY < min_y:
                min_y = minY
            if maxX > max_x:
                max_x = maxX
            if maxY > max_y:
                max_y = maxY
        for gadget_slat_group in gadget_slat_groups:
            # Automatically translating to positive coordinates
            gadget_slat_group.translate((-min_x, -min_y))
        pageWidth = (max_x - min_x + 1) * squareDim
        pageHeight = (max_y - min_y + 1) * squareDim

    if not svg_file_name is None:
        dwg = svgwrite.Drawing(filename=svg_file_name, debug=True)
        dwg.viewbox(width=pageWidth, height=pageHeight)
    if not pdf_file_name is None:
        c = canvas.Canvas(pdf_file_name, pagesize=(pageWidth, pageHeight))

    for gadget_slat_group in gadget_slat_groups:
        if not svg_file_name is None:
            svgGadgetGroup = dwg.add(dwg.g())

        sw_svg_loc = [None, None]
        ne_svg_loc = [None, None]
        sw_pdf_loc = [None, None]
        ne_pdf_loc = [None, None]

        # Ordering of vertical or horizontal first will determine which are on the bottom
        for slats in [gadget_slat_group.vertical_slat_group.slats, gadget_slat_group.horizontal_slat_group.slats]:
            for slat in slats:

                if not svg_file_name is None:
                    svgGroup = svgGadgetGroup.add(dwg.g())

                (curr_x, curr_y) = (slat.begin[0], slat.begin[1])
                while (curr_x, curr_y) <= (slat.end[0], slat.end[1]):

                    fillColor = None

                    if slat.b_horizontal:
                        x_dim = squareDim
                        y_dim = squareDim * skinnyFactor
                        x_offset = 0
                        y_offset = (squareDim - (squareDim * skinnyFactor)) / 2
                        if 'horizontal' in color_mapping:
                            fillColor = color_mapping['horizontal']

                    else:
                        x_dim = squareDim * skinnyFactor
                        y_dim = squareDim
                        x_offset = (squareDim - (squareDim * skinnyFactor)) / 2
                        y_offset = 0
                        if 'vertical' in color_mapping:
                            fillColor = color_mapping['vertical']

                    io_loc = slat.getIOLoc((curr_x, curr_y))
                    if not io_loc is None:
                        if not io_loc[2] is None:
                            fillColor = io_loc[2]
                        elif io_loc[1] in color_mapping:
                            fillColor = color_mapping[io_loc[1]]

                    if (fillColor is None) or (fillColor == defaultFillColor):
                        if not slat.color is None:
                            fillColor = slat.color
                        else:
                            fillColor = defaultFillColor
                    fillColor = fillColor.lower()
                    
                    if not svg_file_name is None:
                        x_loc = (curr_x * squareDim) + x_offset
                        y_loc = pageHeight - ((curr_y + 1) * squareDim) + y_offset

                        if (sw_svg_loc[0] is None) or (x_loc < sw_svg_loc[0]):
                            sw_svg_loc[0] = x_loc
                        if (sw_svg_loc[1] is None) or (y_loc < sw_svg_loc[1]):
                            sw_svg_loc[1] = y_loc
                        if (ne_svg_loc[0] is None) or (x_loc > ne_svg_loc[0]):
                            ne_svg_loc[0] = x_loc
                        if (ne_svg_loc[1] is None) or (y_loc > ne_svg_loc[1]):
                            ne_svg_loc[1] = y_loc

                        svgSquareGroup = svgGroup.add(dwg.g())

                        svgSquareGroup.add(dwg.rect(insert=(x_loc * svgUnits, y_loc * svgUnits),
                                                    size=(x_dim * svgUnits, y_dim * svgUnits),
                                                    fill=fillColor,
                                                    stroke='black',
                                                    stroke_width=strokeWidth))

                        if b_show_labels and (not io_loc is None):
                            text = io_loc[1]
                            if len(text) < 3:
                                text = f' {text} '
                            textLen = len(text)
                            textSize = x_dim / textLen * 0.7
                            textLength = min(x_dim * 0.7, textLen)
                            xOffset = x_loc + (x_dim / 2) - textLength / 2
                            text = dwg.text(text, insert=(xOffset * svgUnits, (y_loc + y_dim * .6) * svgUnits),
                                            textLength=textLength * svgUnits, font_size=textSize * svgUnits,
                                            style=textStyle)
                            svgSquareGroup.add(text)

                    if not pdf_file_name is None:
                        c.setFillColor(fillColor)
                        x_loc = curr_x * squareDim + x_offset
                        y_loc = curr_y * squareDim + y_offset
                        c.rect(x_loc, y_loc, x_dim, y_dim, stroke=strokeWidth, fill=1)

                        if (sw_pdf_loc[0] is None) or (x_loc < sw_pdf_loc[0]):
                            sw_pdf_loc[0] = x_loc
                        if (sw_pdf_loc[1] is None) or (y_loc < sw_pdf_loc[1]):
                            sw_pdf_loc[1] = y_loc
                        if (ne_pdf_loc[0] is None) or (x_loc > ne_pdf_loc[0]):
                            ne_pdf_loc[0] = x_loc
                        if (ne_pdf_loc[1] is None) or (y_loc > ne_pdf_loc[1]):
                            ne_pdf_loc[1] = y_loc

                        if b_show_labels and (not io_loc is None):
                            text = io_loc[1]
                            if len(text) < 3:
                                text = f' {text} '
                            textLen = len(text)
                            fontSize = x_dim / (0.6 * textLen)

                            y_loc = y_loc + y_dim / 3.0
                            x_loc = x_loc + (fontSize * 0.4) #x_loc + x_dim / 4

                            textobject = c.beginText(x_loc, y_loc)
                            textobject.setFont(pdfFontBold, fontSize)
                            textobject.setFillColorRGB(0, 0, 0)
                            textobject.textOut(text)
                            c.drawText(textobject)

                    if slat.b_horizontal:
                        curr_x = curr_x + 1
                    else:
                        curr_y = curr_y + 1
        location_map[gadget_slat_group.gadget_name] = ((sw_svg_loc, ne_svg_loc), (sw_pdf_loc, ne_pdf_loc))

    if tile_set is not None:
        translation = (0,0)
        body_dims = 30
        for tile_type_name in tile_set:
            tile_type = tile_set[tile_type_name]
            pdf_trans = location_map[tile_type_name][1][0]
            svg_trans = (location_map[tile_type_name][0][0][0], location_map[tile_type_name][0][1][1])
            tile_type.MakeImage(pdfCanvas=c, svgCanvas=dwg, svg_trans=svg_trans, pdf_trans=pdf_trans, body_dims=body_dims)

    if not svg_file_name is None:
        dwg.save()
        print(f"Slats saved to SVG file '{svg_file_name}'.")
    if not pdf_file_name is None:
        c.save()
        print(f"Slats saved to PDF file '{pdf_file_name}'.")

def GetTilesAndSlatsHelper(gadget_definition, dictionary_generator, scale_factor, var_dict=None, motif=None):
    all_tiles_dict = dict()
    gadget_slat_groups = []

    dict_list = dictionary_generator(var_dict)

    for slat_dict in dict_list:
        tile_dict = GetGadgetTiles(slat_dict['name'], gadget_definition, slat_dict['label'], slat_dict['color'],
                                   slat_dict)
        combineTileDicts(tile_dict, all_tiles_dict)
        gadget_slat_group = GetGadgetSlats(gadget_definition, scale_factor, slat_dict['name'], slat_dict, motif)
        gadget_slat_groups.append(gadget_slat_group)

    return (all_tiles_dict, gadget_slat_groups)


def MergeMatchingSlats(gadget_slat_groups_list):
    # Check for slats that match at all io_locs except for linking locations
    slat_templates = dict()
    slat_templates_ex = dict()
    domain_set = set()
    for gadget_slat_groups in gadget_slat_groups_list:
        for gadget_slat_group in gadget_slat_groups:
            for slat_list in [gadget_slat_group.vertical_slat_group.slats, gadget_slat_group.horizontal_slat_group.slats]:
                for slat in slat_list:
                    slat_template = []
                    slat_template_ex = []
                    for i in range(slat.getLength()):
                        h = v = 0
                        if slat.b_horizontal:
                            h = i
                        else:
                            v = i
                        coords = (slat.begin[0]+h, slat.begin[1]+v)
                        io_loc = slat.getIOLoc(coords)
                        if not io_loc is None:
                            domain_set.add(io_loc[1])
                            if (len(io_loc) > 3) and (io_loc[3] == True):
                                #print(f'found a crosser at {coords} for slat {slat.name} with label {io_loc[1]}')
                                slat_template.append(True)
                                slat_template_ex.append(io_loc[1])
                            else:
                                slat_template.append(io_loc[1])
                                slat_template_ex.append(io_loc[1])
                        else:
                            slat_template.append(None)
                            slat_template_ex.append(None)
                    #print(f'slat_template: {slat_template}')
                    slat_templates[slat.name] = slat_template
                    slat_templates_ex[slat.name] = slat_template_ex

    replacement_domain_base_name = 'xg'
    replacement_domain_base_num = 1
    slat_removal_list = []
    replacement_domain_mapping = dict()
    for slat_name1 in slat_templates:
        slat_template1 = slat_templates[slat_name1]
        slat_template_ex1 = slat_templates_ex[slat_name1]
        for slat_name2 in slat_templates:
            slat_template2 = slat_templates[slat_name2]
            slat_template_ex2 = slat_templates_ex[slat_name2]
            if (slat_name1 != slat_name2) and (slat_template1 == slat_template2):
                #print(f'found matching slat templates for {slat_name1} and {slat_name2}!')
                # Delete one slat from each matching pair and sync the connection glues
                if (not slat_name1 in slat_removal_list) and (not slat_name2 in slat_removal_list):
                    slat_removal_list.append(slat_name1)
                    for template_loc in range(len(slat_template1)):
                        if slat_template1[template_loc] == True:
                            slat1_domain = slat_template_ex1[template_loc]
                            slat2_domain = slat_template_ex2[template_loc]
                            current_replacement_domain = f'{replacement_domain_base_name}{replacement_domain_base_num}'
                            if current_replacement_domain in domain_set:
                                print(f'ERROR: In MergeMatchingSlats, name collision with replacement glue {current_replacement_domain}')
                            replacement_domain_base_num += 1
                            replacement_domain_mapping[slat1_domain] = current_replacement_domain  # We shouldn't really need to add these, since the slat is being deleted, but what the heck...
                            replacement_domain_mapping[slat2_domain] = current_replacement_domain
                            if slat1_domain[-1] == '*':
                                replacement_domain_mapping[f'{slat2_domain[0:-1]}'] = f'{current_replacement_domain}'
                                replacement_domain_mapping[f'{slat1_domain[0:-1]}'] = f'{current_replacement_domain}'
                            else:
                                replacement_domain_mapping[f'{slat2_domain}*'] = f'{current_replacement_domain}'
                                replacement_domain_mapping[f'{slat1_domain}*'] = f'{current_replacement_domain}'

    #print(f'Merged slat types to remove: {slat_removal_list}')

    deletion_count = 0
    for gadget_slat_groups in gadget_slat_groups_list:
        for gadget_slat_group in gadget_slat_groups:
            for slat_group in [gadget_slat_group.vertical_slat_group, gadget_slat_group.horizontal_slat_group]:
                new_slat_list = []
                comp = '' if (slat_group == gadget_slat_group.vertical_slat_group) else '*'
                for slat in slat_group.slats:
                    if not slat.name in slat_removal_list:
                        new_slat = copy.deepcopy(slat)

                        for i in range(new_slat.getLength()):
                            h = v = 0
                            if slat.b_horizontal:
                                h = i
                            else:
                                v = i
                            coords = (slat.begin[0] + h, slat.begin[1] + v)
                            io_loc = slat.getIOLoc(coords)
                            if io_loc[1] in replacement_domain_mapping:
                                if not io_loc is None:
                                    new_slat.setIOLoc(coords, f'{replacement_domain_mapping[io_loc[1]]}{comp}', io_loc[2], True)

                        new_slat_list.append(new_slat)
                    else:
                        deletion_count += 1
                slat_group.slats = new_slat_list
    print(f'After merging slat types, deleted {deletion_count} duplicate slat types.')
    if deletion_count != len(slat_removal_list):
        print(f'WARNING: not all duplicate slats were deleted in MergeMatchingSlats')



#########################################################
## Functions for working with pools of glue domains (a.k.a. handles)

def_empty_marker = -999


def GetMaximumMatchingSet(seq1: list, seq2: list, matching_limit: int):  # , empty_loc_marker=None):
    # print(f'DEBUG: getting hamming distance of {slat1_seq} and {slat2_seq}')
    if not len(seq1) == len(seq2):
        print(f'WARNING: Different slat lengths in GetMaximumMatchingSetSize: {seq1} and {seq2}')

    ##    empty_marker1 = def_empty_marker
    ##    while (empty_marker1 in slat1_seq) or (empty_marker1 in slat2_seq):
    ##        empty_marker1 = empty_marker1 + empty_marker1
    ##    empty_marker2 = empty_marker1 + empty_marker1
    ##    while (empty_marker2 in slat1_seq) or (empty_marker2 in slat2_seq):
    ##        empty_marker2 = empty_marker2 + empty_marker2
    ##    seq1 = [empty_marker1 if item == empty_loc_marker else item for item in slat1_seq]
    ##    seq2 = [empty_marker2 if item == empty_loc_marker else item for item in slat2_seq]

    pair_num = 0
    for (s1, s2) in [(seq1, seq2), (seq2, seq1), (seq1[::-1], seq2), (seq2, seq1[::-1])]:
        for i in range(len(s1)):
            if i == 0:
                s2_curr = s2
            else:
                s2_curr = s2[:-i]

            # print(f's1: {s1[i:]},\ns2: {s2_curr}\n')
            num_diff = hamming(s1[i:], s2_curr) * len(s1[i:])
            num_same = len(s1[i:]) - num_diff
            if num_same > matching_limit:
                match_set = []
                for j in range(len(s1[i:])):
                    if s1[i:][j] == s2_curr[j]:
                        index = i + j
                        if pair_num == 1:
                            index = j
                        elif pair_num == 2:
                            index = len(seq1) - (i + j) - 1
                        elif pair_num == 3:
                            index = len(seq1) - j - 1
                        match_set.append(index)
                ##                print(f'max matches with:')
                ##                print(f'  {s1[i:]}')
                ##                print(f'  {s2_curr}\n')
                return match_set
        pair_num = pair_num + 1
        # print(f'max_match {max_match}!')
    # print(f'max_match: {max_match}')
    return None


def GetOverlapStatsForPairOfSlats(seq1, seq2, check_rotation=True, include_complementation=False):  # , threshold):
    if not len(seq1) == len(seq2):
        print(f'WARNING: Different slat lengths in GetOverlapStatsForPairOfSlats: {seq1} and {seq2}')

    if include_complementation:
        seq1_has_complements = False
        for domain in seq1:
            if domain.endswith('*'):
                seq1_has_complements = True
        seq2_has_complements = False
        for domain in seq2:
            if domain.endswith('*'):
                seq2_has_complements = True
        if (seq1_has_complements == True) and (seq2_has_complements == False):
            new_seq1 = []
            for domain in seq1:
                if not domain.endswith('*'):
                    print(f'ERROR: in GetOverlapStatsForPairOfSlats, only some domains in list are complements: {seq1}')
                else:
                    new_seq1.append(domain[:-1])
            seq1 = new_seq1
        elif (seq1_has_complements == False) and (seq2_has_complements == True):
            new_seq2 = []
            for domain in seq2:
                if not domain.endswith('*'):
                    print(f'ERROR: in GetOverlapStatsForPairOfSlats, only some domains in list are complements: {seq1}')
                else:
                    new_seq2.append(domain[:-1])
            seq2 = new_seq2

    empty_marker1 = def_empty_marker
    while (empty_marker1 in seq1) or (empty_marker1 in seq2):
        empty_marker1 = empty_marker1 + empty_marker1
    empty_marker2 = empty_marker1 + empty_marker1
    while (empty_marker2 in seq1) or (empty_marker2 in seq2):
        empty_marker2 = empty_marker2 + empty_marker2
    seq1 = [empty_marker1 if item == def_empty_marker else item for item in seq1]
    seq2 = [empty_marker2 if item == def_empty_marker else item for item in seq2]

    mixes = [(seq1, seq2), (seq2, seq1)]
    if check_rotation:
        mixes = mixes + [(seq1[::-1], seq2), (seq2, seq1[::-1])]

    overlaps_dict = dict()
    max_overlap = 0
    pair_num = 0
    max_example = []
    for m in range(len(mixes)):
        (s1, s2) = mixes[m]
        for i in range(len(s1)):
            if i == 0:
                if m < 2:
                    continue
                else:
                    s2_curr = s2
            else:
                s2_curr = s2[:-i]
            num_diff = hamming(s1[i:], s2_curr) * len(s1[i:])
            num_same = int(len(s1[i:]) - num_diff)

            # match_set = []
            # for j in range(len(s1[i:])):
            #     if s1[i:][j] == s2_curr[j]:
            #         index = i + j
            #         if pair_num == 1:
            #             index = j
            #         elif pair_num == 2:
            #             index = len(seq1) - (i + j) - 1
            #         elif pair_num == 3:
            #             index = len(seq1) - j - 1
            #         match_set.append(index)
            # if len(match_set) > 15:
            #     print(f'{len(match_set)} matches with:')
            #     print(f'  {str(s1[i:]).replace(str(empty_marker1), "0")}')
            #     print(f'  {str(s2_curr).replace(str(empty_marker2), "0")}\n')

            if num_same > max_overlap:
                max_overlap = num_same
                if s1 == seq1:
                    max_example = [s1[i:], s2_curr]
                else:
                    max_example = [s2_curr, s1[i:]]

            if num_same in overlaps_dict:
                overlaps_dict[num_same] = overlaps_dict[num_same] + 1
            else:
                overlaps_dict[num_same] = 1

        pair_num = pair_num + 1
    return (max_overlap, overlaps_dict, max_example)


def DoesSlatExceedMatchingLimit(new_slat, existing_slats, matching_limit):
    for existing_slat in existing_slats:
        match_set = GetMaximumMatchingSet(new_slat, existing_slat, matching_limit)
        if not match_set is None:
            return match_set
    return None


# def AssignSlatsFromPool(pool_size, max_k_naum, horizontal_map, vertical_map, horizontal_slat_lists, vertical_slat_lists,
#                         b_allow_handle_reuse=True):
#     handle_pool = [f'h{handle}' for handle in range(pool_size)]
#
#     for slat_list in horizontal_slat_lists:
#         slat_length = len(slat_list)
#         if b_allow_handle_reuse:
#             slat = random.choices(handle_pool, k=slat_length)
#         else:
#             slat = handle_pool.copy()
#             random.shuffle(slat)
#             slat = slat[:slat_length]
#         for i in len(slat_list):
#             # slat_list[i] = slat[i]
#             ####figure this out... horizontal_map[slat_list[i]]...
#             pass


##class CoordsAndHandlesSlatSet:
##    def __init__(self, slat_length):
##        self.coords_map = dict()
##        self.coords_slat = [None * slat_length]
##        self.handles_slat = [None * slat_length]
##
##    def getCoordsSlat(self):
##        return self.coords_slat
##
##    def getHandlesSlat(self):
##        return self.handles_slat
##
##    def setHandleAtCoords(coords, handle):
##        self.handles_slat[self.coords_map[coords]] = handle
##
##    def getHandleAtCoords(coords):
##        return self.handles_slat[self.coords_map[coords]]


def GetCoordToSlatsMaps(gadget_slat_groups):
    horizontal_map = dict()
    vertical_map = dict()
    horizontal_slat_lists = []
    vertical_slat_lists = []

    ##    horizontal_coords = set()
    ##    vertical_coords = set()

    vertical_slat_num = 0
    horizontal_slat_num = 0
    for gadget_slat_group in gadget_slat_groups:
        for slat in gadget_slat_group.horizontal_slat_group.slats:
            horizontal_slat_list = []
            for loc_index in range(slat.getLength()):
                loc = slat.getLocByIndex(loc_index)
                horizontal_map[(horizontal_slat_num, loc_index)] = loc[0]
                horizontal_slat_list.append(loc[0])

            ##                if loc[0] in horizontal_coords:
            ##                    print(f'ERROR: horizontal_coords already contains {loc[0]}')
            ##                horizontal_coords.add(loc[0])

            # print(f'horizontal slat {i}: {horizontal_slat_list}')
            horizontal_slat_lists.append(horizontal_slat_list)
            horizontal_slat_num = horizontal_slat_num + 1

        for slat in gadget_slat_group.vertical_slat_group.slats:
            vertical_slat_list = []
            for j in range(slat.getLength()):
                loc = slat.getLocByIndex(j)
                vertical_map[loc[0]] = (vertical_slat_num, j)
                vertical_slat_list.append(loc[0])

            ##                if loc[0] in vertical_coords:
            ##                    print(f'ERROR: vertical_coords already contains {loc[0]}')
            ##                vertical_coords.add(loc[0])

            vertical_slat_lists.append(vertical_slat_list)
            vertical_slat_num = vertical_slat_num + 1

    ##    hd = copy.deepcopy(horizontal_coords)
    ##    hd.difference_update(vertical_coords)
    ##    print(f'horizontal difference: {hd}')
    ##
    ##    vd = copy.deepcopy(vertical_coords)
    ##    vd.difference_update(horizontal_coords)
    ##    print(f'vertical difference: {vd}')

    return (horizontal_map, vertical_map, horizontal_slat_lists, vertical_slat_lists)


def NullifyLocationsWithoutHandles(horizontal_slat_lists, vertical_slat_lists, horizontal_null=-1, vertical_null=-2):
    for horizontal_slat_list in horizontal_slat_lists:
        for i in range(len(horizontal_slat_list)):
            if type(horizontal_slat_list[i]) == tuple:
                horizontal_slat_list[i] = horizontal_null

    for vertical_slat_list in vertical_slat_lists:
        # v = copy.deepcopy(vertical_slat_list)
        for i in range(len(vertical_slat_list)):
            if type(vertical_slat_list[i]) == tuple:
                print(f'WARNING: unmatched coordinate of vertical slat: {vertical_slat_list[i]}')
                vertical_slat_list[i] = vertical_null


# GetMutuallyExclusiveSlatGroupFromHandlePool is used to generate a group of slats with domains
# from a handle pool such that the max allowed k_num between any two slats is max_k_num
def GetMutuallyExclusiveSlatGroupFromHandlePool(pool_size, num_slats, slat_length, max_k_num, b_allow_handle_reuse=True,
                                                max_attempts=5000, report_increment=200):
    handle_slat_lists = []

    if not pool_size is None:

        if not max_k_num is None:
            print(f'Maximum allowable glue matches between slats will be {max_k_num}')
            max_allowable_matches = max_k_num
        else:
            print(f'There will be no restriction on the number of glue matches between slats')
            max_allowable_matches = slat_length

        handle_pool = [handle for handle in range(pool_size)]

        print(
            f'Attempting to generate {num_slats} slats from handle pool... (will time out after {max_attempts} attempts)')
        for attempt in range(max_attempts):

            if b_allow_handle_reuse:
                slat = random.choices(handle_pool, k=slat_length)
            else:
                slat = handle_pool.copy()
                random.shuffle(slat)
                slat = slat[:slat_length]
            # print(f'slat {len(slats)}: {slat}')

            if not DoesSlatExceedMatchingLimit(slat, handle_slat_lists, max_allowable_matches):
                handle_slat_lists.append(slat)
                if len(handle_slat_lists) == num_slats:
                    break
            if (attempt % report_increment) == 0:
                print(f'  After attempt {attempt}, num slats: {len(handle_slat_lists)}')

        if len(handle_slat_lists) == num_slats:
            print(f'Successfully generated {len(handle_slat_lists)} slats from handle pool after {attempt} attempts.')
        else:
            print(
                f'Timed out generating slats from handle pool after {attempt} attempts. Generated {len(handle_slat_lists)} of {num_slats} desired slats.')
            return False

    return handle_slat_lists


def ApplySlatHandlePool(handle_slat_lists, gadget_slat_groups, vertical_null):
    # Assign handle sets to horizontal slats, keeping track of coorinates swapped out.
    # Then, assign the complements to the matching vertical domains.
    coord_to_handle_map = dict()
    hor_slat_num = 0
    for gadget_slat_group in gadget_slat_groups:
        for slat in gadget_slat_group.horizontal_slat_group.slats:

            handle_slat = handle_slat_lists[hor_slat_num]
            for j in range(slat.getLength()):
                loc = slat.getLocByIndex(j)
                handle = f'{handle_slat[j]}'
                slat.setIOLoc(loc[0], handle)
                coord_to_handle_map[loc[0]] = handle

            hor_slat_num = hor_slat_num + 1

    for gadget_slat_group in gadget_slat_groups:
        for slat in gadget_slat_group.vertical_slat_group.slats:
            for j in range(slat.getLength()):
                loc = slat.getLocByIndex(j)
                if loc[0] in coord_to_handle_map:
                    handle = f'{coord_to_handle_map[loc[0]]}*'
                    slat.setIOLoc(loc[0], handle)
                else:
                    slat.setIOLoc(loc[0], f'{vertical_null}')
                    # print(f'WARNING: coords {loc[0]} not found in coord_to_handle_map.')
    return True


def GetKNumDistribution(gadget_slat_groups):
    pass


def GetSlatsAsDomainLists(slat_types_dict: dict) -> dict:
    slat_types_domain_lists_dict = dict()
    for slat_type_name in slat_types_dict:
        domain_list = slat_types_dict[slat_type_name].get_domain_label_list()
        slat_types_domain_lists_dict[slat_type_name] = domain_list
    return slat_types_domain_lists_dict

def GetIncorrectBindingStatsForSlats(horiz_slat_types_domain_lists_dict, vert_slat_types_domain_lists_dict):

    half_self_check_max_overlap = 0
    half_self_check_max_overlap_count = 0
    max_half_self_check_example = None
    self_check_max_overlap = 0
    self_check_max_overlap_count = 0
    max_self_check_example = None
    for slat_type_pair in [(horiz_slat_types_domain_lists_dict, horiz_slat_types_domain_lists_dict), (vert_slat_types_domain_lists_dict, vert_slat_types_domain_lists_dict)]:
        for slat_type1_name in slat_type_pair[0]:
            for slat_type2_name in slat_type_pair[1]:
                slat_type1 = slat_type_pair[0][slat_type1_name]
                slat_type2 = slat_type_pair[1][slat_type2_name]
                (max_overlap, overlaps_dict, max_example) = GetOverlapStatsForPairOfSlats(slat_type1, slat_type2, check_rotation=True,
                                                                                          include_complementation=False)
                if max_overlap > self_check_max_overlap:
                    self_check_max_overlap = max_overlap
                    self_check_max_overlap_count = 1
                    max_self_check_example = [slat_type1, slat_type2] + max_example
                elif max_overlap == self_check_max_overlap:
                    self_check_max_overlap_count += 1


                st1 = slat_type1[0:int(len(slat_type1)/2)]
                st2 = slat_type2[0:int(len(slat_type2)/2)]
                (max_overlap, overlaps_dict, max_example) = GetOverlapStatsForPairOfSlats(st1, st2, check_rotation=True,
                                                                                          include_complementation=False)
                if max_overlap > half_self_check_max_overlap:
                    half_self_check_max_overlap = max_overlap
                    half_self_check_max_overlap_count = 1
                    max_half_self_check_example = [slat_type1, slat_type2] + max_example
                elif max_overlap == half_self_check_max_overlap:
                    half_self_check_max_overlap_count += 1

    cross_check_max_overlap = 0
    cross_check_max_overlap_count = 0
    max_cross_check_example = None
    for slat_type1_name in horiz_slat_types_domain_lists_dict:
        for slat_type2_name in vert_slat_types_domain_lists_dict:
            slat_type1 = horiz_slat_types_domain_lists_dict[slat_type1_name]
            slat_type2 = vert_slat_types_domain_lists_dict[slat_type2_name]
            (max_overlap, overlaps_dict, max_example) = GetOverlapStatsForPairOfSlats(slat_type1, slat_type2, check_rotation=True,
                                                                                      include_complementation=True)
            if max_overlap > cross_check_max_overlap:
                cross_check_max_overlap = max_overlap
                cross_check_max_overlap_count = 1
                max_cross_check_example = [slat_type1, slat_type2] + max_example
            elif max_overlap == cross_check_max_overlap:
                cross_check_max_overlap_count += 1

    results_dict = {'max_half_self_overlap': half_self_check_max_overlap, 'max_self_overlap': self_check_max_overlap, 'max_cross_overlap': cross_check_max_overlap,
                    'max_half_self_overlap_count': half_self_check_max_overlap_count, 'max_self_overlap_count': self_check_max_overlap_count, 'max_cross_overlap_count': cross_check_max_overlap_count,
                    'max_half_self_overlap_example': max_half_self_check_example, 'max_self_overlap_example': max_self_check_example, 'max_cross_overlap_example': max_cross_check_example}
    return results_dict

def AssignSlatsFromPool(horiz_slat_types, vert_slat_types, num_handles, handle_pool=None, max_error=None, max_iterations=100, check_rotation=False, force_max_error=False):
    horiz_domain_set = set()
    for slat_name in horiz_slat_types:
        slat_type = horiz_slat_types[slat_name]
        horiz_domain_set = horiz_domain_set.union(set(slat_type.get_domain_label_list()))
    print(f'Attempting to assign {num_handles} handles to slats with {len(horiz_domain_set)} domains...')

    if not handle_pool:
        handle_pool = [f'h{i}' for i in range(num_handles)]

    horiz_slat_types_ex = GetSlatsAsDomainLists(horiz_slat_types)
    vert_slat_types_ex = GetSlatsAsDomainLists(vert_slat_types)

    domain_set = set()
    for horiz_slat_type in horiz_slat_types_ex:
        for domain in horiz_slat_types_ex[horiz_slat_type]:
            domain_set.add(domain)

    domain_to_handle_map = dict()
    do_assignment = True
    iteration_num = 0
    while do_assignment:    # potentially loop if there is a maximum value for erroneous binding set

        domain_list = list(domain_set)
        count = 0
        for domain in domain_list:
            if count < len(handle_pool):
                domain_to_handle_map[domain] = handle_pool[count]
                count += 1
            else:
                domain_to_handle_map[domain] = handle_pool[random.randrange(num_handles)]
            domain_to_handle_map[f'{domain}*'] = domain_to_handle_map[domain]
        #print(f'domain_to_handle_map: {domain_to_handle_map}')

        handle_marker = ''
        horiz_slat_types_ex2 = copy.deepcopy(horiz_slat_types_ex)
        vert_slat_types_ex2 = copy.deepcopy(vert_slat_types_ex)

        for slat_types in [horiz_slat_types_ex2, vert_slat_types_ex2]:
            for slat_type_name in slat_types:
                slat_type = slat_types[slat_type_name]
                #print(f'slat_type: {slat_type}, slat_types: {slat_types}')
                for i in range(len(slat_type)):
                    slat_types[slat_type_name][i] = f'{domain_to_handle_map[slat_type[i]]}{handle_marker}'
            handle_marker = '*'

        check_results_dict = GetIncorrectBindingStatsForSlats(horiz_slat_types_ex2, vert_slat_types_ex2)
        max_half_self_overlap = check_results_dict['max_half_self_overlap']
        max_cross_overlap = check_results_dict['max_cross_overlap']

        iteration_num += 1
        if iteration_num > max_iterations:
            do_assignment = False
        elif (max_half_self_overlap > max_error) or (max_cross_overlap > max_error):
            do_assignment = True
        else:
            do_assignment = False
    if (max_half_self_overlap > max_error) or (max_cross_overlap > max_error):
        print(f'ERROR: After maximum allowed iterations ({iteration_num}) maximum errors still exceeded: max_half_self_overlap = {max_half_self_overlap}, max_cross_overlap = {max_cross_overlap}')
    else:
        print(f'Handle assignment succeeded.')

    handle_marker = ''
    for slat_types in [horiz_slat_types, vert_slat_types]:
        for slat_type_name in slat_types:
            slat_type = slat_types[slat_type_name]
            for block in slat_type.blocks:
                if block in slat_type.domains:
                    domains = slat_type.domains[block]
                    domain_list = []
                    for domain in domains:
                        new_domain = (f'{domain_to_handle_map[domain[0]]}{handle_marker}', domain[1], domain[2])
                        domain_list.append(new_domain)
                    slat_type.domains[block] = domain_list
        handle_marker = '*'

    return (horiz_slat_types, vert_slat_types)


def is_subsequence(subsequence, sequence):
    shift_range = len(sequence)-len(subsequence)+1
    for i in range(shift_range):
        if sequence[i:i+len(subsequence)] == subsequence:
            return i
    return -1

def GetShiftedSlatsForOverlap(example_slat1, example_subslat1, example_slat2, example_subslat2):
    noncomp_slat1 = copy.deepcopy(example_slat1)
    noncomp_slat2 = copy.deepcopy(example_slat2)
    slat1_flip = False
    slat2_flip = False
    if ('*' in noncomp_slat1[0]):
        if (not '*' in noncomp_slat2[0]):
            noncomp_slat1 = [a[:-1] for a in noncomp_slat1]
    elif ('*' in noncomp_slat2[0]):
        noncomp_slat2 = [a[:-1] for a in noncomp_slat2]

    b_error = False
    shifted_slat1 = copy.deepcopy(example_slat1)
    slat1_shift = is_subsequence(example_subslat1, noncomp_slat1)
    #print(f'slat1_shift: {slat1_shift}, example_subslat1: {example_subslat1}, noncomp_slat1: {noncomp_slat1}')
    if slat1_shift == -1:
        slat1_shift = is_subsequence(example_subslat1, noncomp_slat1[::-1])
        shifted_slat1 = shifted_slat1[::-1]
        slat1_flip = True
    if slat1_shift == -1:
        print(f'ERROR: no shift found for {example_subslat1} {example_slat1}')
        b_error = True

    shifted_slat2 = copy.deepcopy(example_slat2)
    slat2_shift = is_subsequence(example_subslat2, noncomp_slat2)
    #print(f'slat2_shift: {slat2_shift}, example_subslat2: {example_subslat2}, noncomp_slat2: {noncomp_slat2}')
    if slat2_shift == -1:
        slat2_shift = is_subsequence(example_subslat2, noncomp_slat2[::-1])
        shifted_slat2 = shifted_slat2[::-1]
        slat2_flip = True
    if slat2_shift == -1:
        print(f'ERROR: no shift found for {example_subslat2} {shifted_slat2}')
        b_error = True

    if not b_error:
        shifted_slat1 = ['___' for i in range(slat1_shift+slat2_shift)] + shifted_slat1
        # print(f'shift: {shift} for {max_self_example_slat1}:\n{shifted_slat1}')
        shifted_slat2 = shifted_slat2 + ['___' for i in range(slat1_shift + slat2_shift)]
        # print(f'shift: {shift} for {max_self_example_slat2}:\n{shifted_slat2}')
    return (slat1_shift, 0-slat2_shift, slat1_flip, slat2_flip, shifted_slat1, shifted_slat2)


def GetExampleErrorSystem(horiz_slat_types, vert_slat_types, results_dict):

    max_self_overlap_example = results_dict['max_self_overlap_example']
    max_self_example_slat1 = max_self_overlap_example[0]
    max_self_example_slat2 = max_self_overlap_example[1]
    max_self_example_subslat1 = max_self_overlap_example[2]
    max_self_example_subslat2 = max_self_overlap_example[3]
    (max_self_example_slat1_shift, max_self_example_slat2_shift,
     max_self_example_slat1_flip, max_self_example_slat2_flip,
     shifted_max_self_slat1, shifted_max_self_slat2) = GetShiftedSlatsForOverlap(max_self_example_slat1,
                                                                                 max_self_example_subslat1,
                                                                                 max_self_example_slat2,
                                                                                 max_self_example_subslat2)

    max_half_self_overlap_example = results_dict['max_half_self_overlap_example']
    max_half_self_example_slat1 = max_half_self_overlap_example[0]
    max_half_self_example_slat2 = max_half_self_overlap_example[1]
    max_half_self_example_subslat1 = max_half_self_overlap_example[2]
    max_half_self_example_subslat2 = max_half_self_overlap_example[3]
    (max_half_self_example_slat1_shift, max_half_self_example_slat2_shift,
     max_half_self_example_slat1_flip, max_half_self_example_slat2_flip,
     shifted_max_half_self_slat1, shifted_max_half_self_slat2) = GetShiftedSlatsForOverlap(max_half_self_example_slat1,
                                                                                           max_half_self_example_subslat1,
                                                                                           max_half_self_example_slat2,
                                                                                           max_half_self_example_subslat2)

    max_cross_overlap_example = results_dict['max_cross_overlap_example']
    max_cross_example_slat1 = max_cross_overlap_example[0]
    max_cross_example_slat2 = max_cross_overlap_example[1]
    max_cross_example_subslat1 = max_cross_overlap_example[2]
    max_cross_example_subslat2 = max_cross_overlap_example[3]
    (max_cross_example_slat1_shift, max_cross_example_slat2_shift,
     max_cross_example_slat1_flip, max_cross_example_slat2_flip,
     shifted_max_cross_slat1, shifted_max_cross_slat2) = GetShiftedSlatsForOverlap(max_cross_example_slat1,
                                                                                   max_cross_example_subslat1,
                                                                                   max_cross_example_slat2,
                                                                                   max_cross_example_subslat2)

    #print( f'best motif: {results_dict}')  # ,\nmax self overlap example: {max_self_overlap_example}\nmax cross overlap example: {max_cross_overlap_example}')
    #print(f'(max_cross_example_slat1_shift, max_cross_example_slat2_shift, max_cross_example_slat1_flip, max_cross_example_slat2_flip, shifted_max_cross_slat1, shifted_max_cross_slat2) = {(max_cross_example_slat1_shift, max_cross_example_slat2_shift, max_cross_example_slat1_flip, max_cross_example_slat2_flip, shifted_max_cross_slat1, shifted_max_cross_slat2)}')

    slat_system_errors = SlatSystem()
    for slat_types in [horiz_slat_types, vert_slat_types]:
        for slat_type in slat_types:
            slat_system_errors.AddPolyominoType(slat_types[slat_type])

    max_self_example_slat1_name = None
    max_self_example_slat2_name = None
    max_half_self_example_slat1_name = None
    max_half_self_example_slat2_name = None
    max_cross_example_slat1_name = None
    max_cross_example_slat2_name = None
    for slat_type_set in [horiz_slat_types, vert_slat_types]:
        for slat_type_name in slat_type_set:
            slat_type = slat_type_set[slat_type_name]
            slat_type_domain_list = slat_type.get_domain_label_list()
            if slat_type_domain_list == max_self_example_slat1:
                max_self_example_slat1_name = slat_type.name
            if slat_type_domain_list == max_self_example_slat2:
                max_self_example_slat2_name = slat_type.name
            if slat_type_domain_list == max_half_self_example_slat1:
                max_half_self_example_slat1_name = slat_type.name
            if slat_type_domain_list == max_half_self_example_slat2:
                max_half_self_example_slat2_name = slat_type.name
            if slat_type_domain_list == max_cross_example_slat1:
                max_cross_example_slat1_name = slat_type.name
            if slat_type_domain_list == max_cross_example_slat2:
                max_cross_example_slat2_name = slat_type.name
    if max_self_example_slat1_name is None:
        print(f'ERROR: failed to find max_self_example_slat1')
    if max_self_example_slat2_name is None:
        print(f'ERROR: failed to find max_self_example_slat2')
    if max_half_self_example_slat1_name is None:
        print(f'ERROR: failed to find max_half_self_example_slat1')
    if max_half_self_example_slat2_name is None:
        print(f'ERROR: failed to find max_half_self_example_slat2')
    if max_cross_example_slat1_name is None:
        print(f'ERROR: failed to find max_cross_example_slat1')
    if max_cross_example_slat2_name is None:
        print(f'ERROR: failed to find max_cross_example_slat2')
    #print(
    #    f'Max self example slats: {max_self_example_slat1_name} {max_self_example_slat2_name}, max half self example slats {max_half_self_example_slat1_name} {max_half_self_example_slat2_name}, max cross example slats: {max_cross_example_slat1_name} {max_cross_example_slat2_name}')

    # print(f'shifts: {max_self_example_slat1_shift} {max_self_example_slat2_shift} {max_cross_example_slat1_shift} {max_cross_example_slat2_shift}')
    # Make a system showing the example max errors
    max_self_example_slat1_inst = TASSlat(slat_system_errors.GetPolyominoTypeByName(max_self_example_slat1_name))
    max_self_example_slat2_inst = TASSlat(slat_system_errors.GetPolyominoTypeByName(max_self_example_slat2_name))
    max_half_self_example_slat1_inst = TASSlat(slat_system_errors.GetPolyominoTypeByName(max_half_self_example_slat1_name))
    max_half_self_example_slat2_inst = TASSlat(slat_system_errors.GetPolyominoTypeByName(max_half_self_example_slat2_name))
    max_cross_example_slat1_inst = TASSlat(slat_system_errors.GetPolyominoTypeByName(max_cross_example_slat1_name))
    max_cross_example_slat2_inst = TASSlat(slat_system_errors.GetPolyominoTypeByName(max_cross_example_slat2_name))
    if max_self_example_slat1_inst.slatType.IsHorizontal() == False:
        max_self_example_slat1_inst = TASSlat(max_self_example_slat1_inst.slatType.GetRotatedVersion(270))
        max_self_example_slat1_shift = 0 - max_self_example_slat1_shift
    if max_self_example_slat2_inst.slatType.IsHorizontal() == False:
        max_self_example_slat2_inst = TASSlat(max_self_example_slat2_inst.slatType.GetRotatedVersion(270))
        max_self_example_slat1_shift = 0 - max_self_example_slat1_shift

    if max_half_self_example_slat1_inst.slatType.IsHorizontal() == False:
        max_half_self_example_slat1_inst = TASSlat(max_half_self_example_slat1_inst.slatType.GetRotatedVersion(270))
        max_half_self_example_slat1_shift = 0 - max_half_self_example_slat1_shift
    if max_half_self_example_slat2_inst.slatType.IsHorizontal() == False:
        max_half_self_example_slat2_inst = TASSlat(max_half_self_example_slat2_inst.slatType.GetRotatedVersion(270))
        max_half_self_example_slat1_shift = 0 - max_half_self_example_slat1_shift

    if max_cross_example_slat1_inst.slatType.IsHorizontal() == False:
        max_cross_example_slat1_inst = TASSlat(max_cross_example_slat1_inst.slatType.GetRotatedVersion(270))
        max_cross_example_slat1_shift = 0 - max_cross_example_slat1_shift
    if max_cross_example_slat2_inst.slatType.IsHorizontal() == False:
        max_cross_example_slat2_inst = TASSlat(max_cross_example_slat2_inst.slatType.GetRotatedVersion(270))
        max_cross_example_slat1_shift = 0 - max_cross_example_slat1_shift

    if max_self_example_slat1_flip == True:
        max_self_example_slat1_inst = TASSlat(max_self_example_slat1_inst.slatType.GetRotatedVersion(180))
    if max_self_example_slat2_flip == True:
        max_self_example_slat2_inst = TASSlat(max_self_example_slat2_inst.slatType.GetRotatedVersion(180))
    if max_half_self_example_slat1_flip == True:
        max_half_self_example_slat1_inst = TASSlat(max_half_self_example_slat1_inst.slatType.GetRotatedVersion(180))
    if max_half_self_example_slat2_flip == True:
        max_half_self_example_slat2_inst = TASSlat(max_half_self_example_slat2_inst.slatType.GetRotatedVersion(180))
    if max_cross_example_slat1_flip == True:
        max_cross_example_slat1_inst = TASSlat(max_cross_example_slat1_inst.slatType.GetRotatedVersion(180))
    if max_cross_example_slat2_flip == True:
        max_cross_example_slat2_inst = TASSlat(max_cross_example_slat2_inst.slatType.GetRotatedVersion(180))

    max_self_example_slat1_inst.translate_reference_point_to((max_self_example_slat1_shift, 0, 0))
    max_self_example_slat2_inst.translate_reference_point_to((max_self_example_slat2_shift, 1, 0))
    max_half_self_example_slat1_inst.translate_reference_point_to((max_half_self_example_slat1_shift, 3, 0))
    max_half_self_example_slat2_inst.translate_reference_point_to((max_half_self_example_slat2_shift, 4, 0))
    max_cross_example_slat1_inst.translate_reference_point_to((max_cross_example_slat1_shift, 6, 0))
    max_cross_example_slat2_inst.translate_reference_point_to((max_cross_example_slat2_shift, 7, 0))

    slat_system_errors.seed.AddPolyominoForViewingOnly(max_self_example_slat1_inst)
    slat_system_errors.seed.AddPolyominoForViewingOnly(max_self_example_slat2_inst)
    slat_system_errors.seed.AddPolyominoForViewingOnly(max_half_self_example_slat1_inst)
    slat_system_errors.seed.AddPolyominoForViewingOnly(max_half_self_example_slat2_inst)
    slat_system_errors.seed.AddPolyominoForViewingOnly(max_cross_example_slat1_inst)
    slat_system_errors.seed.AddPolyominoForViewingOnly(max_cross_example_slat2_inst)

    return slat_system_errors
