from enum import Enum
import copy
import sys
import getopt
import os
import math
from PyTASTileSystem import Glue, PyTASTileType, PyTASTileSystem, CELLSIZE
from slat_generator import GadgetSlatGroup, SlatGroup, LibSlat, DrawSlatsToFiles, GetTASSlatTypeFromLibSlatType
from SlatSystem import *


# ht:o:d:n:c:m:r:l:xsp
# ["help", "tdp=", "out=", "dir=", "scale=", "count=", "motif=", "color=", "labels=", "xml", "svg", "pdf"]
def DisplayHelp():
    print("\nThis script takes as input an aTAM tile assembly system and outputs a slat system that simulates it.")
    print("There are several caveats, TBD.")
    print("\nUsage:")
    print("-h, --help\t\t\tDisplays this help message.")
    print("-i <name> --input=<name>\tThe name of the Python file containing the macrotile definition functions and maps.")
    print("-t <name>, --tdp=<name>\t\tThe name of the TDP file containing the name of the TDS file, temperature, and")
    print("\t\t\t\t definition of the seed for the aTAM system to be simulated.")
    print("-f, --full\t\t\t[Optional] If this option is used, no input tdp file is required and instead tile")
    print("\t\t\t\t types with all combinations of inputs/outputs supported by the given macrotile maps")
    print("\t\t\t\t will be generated and converted.")
    print("-o <name>, --out=<name>\t\t[Optional] Base name of the output files. If not specified, the base name of the")
    print("\t\t\t\t input tile assembly system will be used.")
    print("-d <dir>, --dir=<dir>\t\t[Optional] Directory for output files. If not specified,")
    print("\t\t\t\t the directory of the input file will be used.")
    print("-n <scale>, --scale=<scale>\t[Optional] Size of slat macrotiles. If not specified, defaults to 4.")
#    print("-c <count>, --count=<count>\t[Optional] Slat count (i.e. number of unique slats per layer")
#    print("\t\t\t\t per macrotile). Defaults to the scale factor value.")
#    print("-m <motif>, --motif=<motif>\t[Optional] If specified, the pattern of slats in each macrotile will be that of")
#    print("\t\t\t\t the specified motif, e.g. 'abbaba'.")
#    print("-r <t/m>, --color=<t/m>\t\t[Optional] If this option is set to 'm', slats will be colored based on the motif")
#    print("\t\t\t\t pattern. Otherwise, they will be colored to match the tiles.")
    print("-l <T/F>, --labels=<T/F>\t[Optional] Show lables in PDF/SVG, T for true, F for False (default T)")
    print("-x, --xml\t\t\t[Optional] If specified, slats will be output to a XML file.")
    print("-s, --svg\t\t\t[Optional] If specified, slats will be output to an SVG file.")
    print("-p, --pdf\t\t\t[Optional] If specified, slats will be output to a PDF file.")
    print("")
    exit()
    
    
N = North = 0
E = East = 1
S = South = 2
W = West = 3
U = Up = 4
D = Down = 5

Dirs = [N,E,S,W]
DirAbbrevArray = ("N", "E", "S", "W", "U", "D")
DirNameMap = {N:"NORTH", E:"EAST", S:"SOUTH", W:"WEST", U:"UP", D:"DOWN"}
DIR_TO_STR = {N:'N', E:'E', S:'S', W:'W', U:'U', D:'D'}

INPUT_MARKERS = {N: 'v', E: '<', S: '^', W: '>'}
OUTPUT_MARKERS = {N: '^', E: '>', S: 'v', W: '<'}

class Orient(Enum):
    HORIZONTAL = "Horizontal"
    HORIZ = "Horizontal"
    H = "Horizontal"
    VERTICAL = "Vertical"
    VERT = "Vertical"
    V = "Vertical"


#####################################################
# Helper functions used for io_conventions
def odd(input):
    if (input % 2) == 1:
        return True
    else:
        return False
def even(input):
    return not odd(input)

def even_and_odd(input):
    return True



def apply_CW_rotation_to_direction(direction, rotation_degrees):
    rotation_count = int(rotation_degrees/90)
    return (direction + rotation_count) % 4
    
def apply_CW_rotation_to_direction_list(direction_list, rotation_degrees):
    rotated_direction_list = []
    for direction in direction_list:
        rotated_direction = apply_CW_rotation_to_direction(direction, rotation_degrees)
        rotated_direction_list.append(rotated_direction)
    return rotated_direction_list

def rotate_coords_CW(coords, degrees_rotation):
    rotated_x = coords[0]
    rotated_y = coords[1]
    for i in range(int(degrees_rotation/90)):
        (rotated_x, rotated_y) = (rotated_y, -rotated_x)
    return ((rotated_x, rotated_y))
    
def rotate_coord_list_CW(coord_list, degrees_rotation):
    rotated_coord_list = []
    for coords in coord_list:
        rotated_coords = rotate_coords_CW(coords, degrees_rotation)
        rotated_coord_list.append(rotated_coords)
    rotated_coord_list.sort()
    return rotated_coord_list

def rotate_gadget_definition_CW(gadget_definition, degrees_rotation):
    gadget_definition = copy.deepcopy(gadget_definition)
    
    for gadget_entry in gadget_definition:

        # coords
        if 'coords' in gadget_entry:
            rotated_coords = rotate_coord_list_CW(gadget_entry['coords'], degrees_rotation)
            gadget_entry['coords'] = rotated_coords

        # outputs
        if 'outputs' in gadget_entry:
            rotated_outputs = []
            for output in gadget_entry['outputs']:
                coords = output[0]
                direction = output[1]
                rotated_outputs.append((rotate_coords_CW(coords, degrees_rotation), apply_CW_rotation_to_direction(direction, degrees_rotation)))
            gadget_entry['outputs'] = rotated_outputs

        # inputs
        if 'inputs' in gadget_entry:
            rotated_inputs = []
            for input_def in gadget_entry['inputs']:
                coords = input_def[0]
                direction = input_def[1]
                rotated_inputs.append((rotate_coords_CW(coords, degrees_rotation), apply_CW_rotation_to_direction(direction, degrees_rotation)))
            gadget_entry['inputs'] = rotated_inputs
                                        
        # active_slats
        if 'active_slats' in gadget_entry:
            active_slats_args = gadget_entry['active_slats'][1]
            rotated_active_slats_args = []
            for arg in active_slats_args:
                if arg in [N,E,S,W]: #TODO: This is a bit dangerous since the drections are just defined as ints...
                    arg = apply_CW_rotation_to_direction(arg, degrees_rotation)
                rotated_active_slats_args.append(arg)
            gadget_entry['active_slats'] = (gadget_entry['active_slats'][0], tuple(rotated_active_slats_args))

        # orientation
        if 'orientation' in gadget_entry:
            if (degrees_rotation % 180) == 90:
                gadget_entry['orientation'] = Orient.V if gadget_entry['orientation'] == Orient.H else Orient.H

        # generic glues
        if 'generic_glues' in gadget_entry:
            generic_glues = gadget_entry['generic_glues']
            rotated_generic_glues = []
            for generic_glue in generic_glues:
                coords = generic_glue[0]
                direction = generic_glue[1]
                glue_label = generic_glue[2]
                rotated_generic_glue = (rotate_coords_CW(coords, degrees_rotation), apply_CW_rotation_to_direction(direction, degrees_rotation), glue_label)
                rotated_generic_glues.append(rotated_generic_glue)
            gadget_entry['generic_glues'] = rotated_generic_glues

    return gadget_definition

def get_CW_most_direction(direction_set: set) -> int:
    if len({N,E,S,W}.union(direction_set)) > 4:
        print(f'ERROR: invalid direction_set {direction_set} passed to get_CW_most_direction.')
        exit()
    elif len(direction_set) > 3:
        print(f'ERROR: invalid direction_set {direction_set} passed to get_CW_most_direction: too long')
        exit()
    elif len(direction_set) < 1:
        print(f'ERROR: empty direction_set passed to get_CW_most_direction')
        exit()
    elif len(direction_set) == 1:
        return list(direction_set)[0]
    elif len(direction_set) == 3:
        missing_dir = list({N,E,S,W}.symmetric_difference(direction_set))[0]
        return (missing_dir - 1) % 4
    else:
        direction_list = list(direction_set)
        if len({W,N}.symmetric_difference(direction_set)) == 0:
            return N
        elif abs(direction_list[0] - direction_list[1]) != 1:
            print(f'ERROR: directions in direction_set {direction_set} are not consecutive.')
            exit()
        else:
            return max(direction_list[0], direction_list[1])

def get_signature_for_side(direction, tile_type):
    strength = tile_type.sides[direction].strength
    if direction in get_tile_input_sides(tile_type):
        return ('in', strength)
    elif direction in get_tile_output_sides(tile_type):
        return ('out', strength)
    else:
        return (None, 0)

def get_key_for_side(direction, tile_type):
    strength = tile_type.sides[direction].strength
    if direction in get_tile_input_sides(tile_type):
        return (direction, 'in', strength)
    elif direction in get_tile_output_sides(tile_type):
        return (direction, 'out', strength)
    else:
        return None
    
def directions_are_adjacent(dir1, dir2):
    if abs(dir1 - dir2) in [1,3]:
        return True
    else:
        return False

def is_horizontal(coord_list):
    if (coord_list[-1][1] - coord_list[0][1]) != 0:
        return False
    elif (coord_list[-1][0] - coord_list[0][0]) != 0:
        return True
    else:
        return False

def is_vertical(coord_list):
    if (coord_list[-1][0] - coord_list[0][0]) != 0:
        return False
    elif (coord_list[-1][1] - coord_list[0][1]) != 0:
        return True
    else:
        return False

def get_overlaps(coord_list1, coord_list2):
    overlaps = []
    for coord in coord_list1:
        if coord in coord_list2:
            overlaps.append(coord)
    return overlaps   

def get_points_from_ends(coords_start, coords_end, b_sort=False):
    x_diff = coords_end[0] - coords_start[0]
    y_diff = coords_end[1] - coords_start[1]

    (x_next, y_next) = (0,0)
    if x_diff < 0:
        x_next = -1
    elif x_diff > 0:
        x_next = 1

    if y_diff < 0:
        y_next = -1
    elif y_diff > 0:
        y_next = 1

    if (x_next != 0) and (y_next != 0):
        print(f"ERROR: Diagonal slat coords {coords_start} to {coords_end}")
        quit()

    coords = []
    coords_curr = coords_start
    while True:
        coords.append(coords_curr)
        if coords_curr == coords_end:
            break
        coords_curr = (coords_curr[0] + x_next, coords_curr[1] + y_next) 

    if b_sort:
        coords.sort()

    return coords

def get_scaled_slat(slat_coords, coop, orientation):
    scaled_slat_coords = []
    for (x,y) in slat_coords:
        for i in range(coop):
            if orientation == Orient.H:
                scaled_slat_coords.append(((x*coop)+i,y*coop))
            elif orientation == Orient.V:
                scaled_slat_coords.append((x*coop,(y*coop)+i))
            else:
                print(f'ERROR: slat is neither horizontal nor vertical: {slat_coords}')
    return scaled_slat_coords

def get_length(slat_coords):
    sorted_slat_coords = copy.deepcopy(slat_coords)
    sorted_slat_coords.sort()
    if is_vertical(slat_coords):
        return slat_coords[-1][1] - slat_coords[0][1] + 1
    elif is_horizontal(slat_coords):
        return slat_coords[-1][0] - slat_coords[0][0] + 1
    else:
        print(f'ERROR: Invalid slat_coords {slat_coords} in get_length')
        quit()
        
def get_tile_input_sides(io_tile_type):
    input_sides = []
    for side_dir in io_tile_type.sides:
        if io_tile_type.sides[side_dir].label.startswith(INPUT_MARKERS[side_dir]):
            input_sides.append(side_dir)
    return input_sides

def get_tile_output_sides(io_tile_type):
    output_sides = []
    for side_dir in io_tile_type.sides:
        if io_tile_type.sides[side_dir].label.startswith(OUTPUT_MARKERS[side_dir]):
            output_sides.append(side_dir)
    return output_sides

def get_slat_name_from_coord_list(slat_coords):
    name_str = f'({slat_coords[0][0]},{slat_coords[0][1]})({slat_coords[-1][0]},{slat_coords[-1][1]})'
    return name_str

class Macrotile:

    def __init__(self, coop, macrotile_base, macrotile_additions_map, macrotile_rotation,
                 h_glue_suffix, v_glue_suffix, io_conventions, overload_set, slat_name_suffix='',
                 motif=None):
        self.coop = coop
        self.macrotile_base = macrotile_base
        self.macrotile_additions_map = macrotile_additions_map
        self.macrotile_rotation = macrotile_rotation
        self.h_glue_suffix = h_glue_suffix
        self.v_glue_suffix = v_glue_suffix
        self.io_conventions = io_conventions
        self.overload_set = overload_set
        self.motif = motif
        self.color = None
        self.slat_name_suffix = slat_name_suffix
        self.slat_color_map = {}

        if (motif is not None) and (len(motif) != coop):
            print(f"ERROR: Length of motif '{motif}' is {len(motif)}, which is not equal to the cooperativity value {coop}")
            quit()

        self.v_slats = []
        self.h_slats = []

        self.coords_to_glues_map = {}
        self.crossing_coords = []
        self.empty_coords = []

    def add_glues_to_lib_slat(self, lib_slat, slat_coords, tile_name):
        glue_suffix = f'{self.h_glue_suffix}' if is_horizontal(slat_coords) else f'{self.v_glue_suffix}'
        for i in range(lib_slat.getLength()):
            if slat_coords[i] in self.coords_to_glues_map:
                lib_slat.setIOLoc(slat_coords[i], f'{self.coords_to_glues_map[slat_coords[i]]}{glue_suffix}')
            elif slat_coords[i] in self.crossing_coords:
                lib_slat.setIOLoc(slat_coords[i], f'{tile_name}({slat_coords[i][0]},{slat_coords[i][1]}){glue_suffix}')

    def add_output_glues(self, slat_group, coop, glue_strength, glue_label, orientation, glue_marker, outputs):
        for output in outputs:
            if (glue_strength == 2):
                keeper_function = lambda n: True
            else:
                keeper_function = self.io_conventions((output[1], 'out'))
            scaled_coords = (output[0][0] * coop, output[0][1] * coop)
            tossed_locs = []
            for x in range(coop):
                x_coord = scaled_coords[0] + x
                for y in range(coop):
                    y_coord = scaled_coords[1] + y

                    if orientation == Orient.H:
                        if (output[1] not in [E, W]):
                            print(f'ERROR: Must have horizontal output for horizontal slat for slat_group {slat_group}')
                            quit()
                        elif keeper_function(y):
                            self.coords_to_glues_map[(x_coord, y_coord)] = f'{glue_label}{glue_marker}({x},{y})H'
                        else:
                            tossed_locs.append((x_coord, y_coord))
                    else:
                        if (output[1] not in [N, S]):
                            print(f'ERROR: Must have vertical output for vertical slat for slat_group {slat_group}')
                            quit()
                        elif keeper_function(x):
                            self.coords_to_glues_map[(x_coord, y_coord)] = f'{glue_label}{glue_marker}({x},{y})V'
                        else:
                            tossed_locs.append((x_coord, y_coord))

            if orientation == Orient.H:
                new_h_slats = []
                for h_slat in self.h_slats:
                    new_h_slat = [(x, y) for (x, y) in h_slat if (x, y) not in tossed_locs]
                    new_h_slats.append(new_h_slat)
                self.h_slats = new_h_slats
            else:
                new_v_slats = []
                for v_slat in self.v_slats:
                    new_v_slat = [(x, y) for (x, y) in v_slat if (x, y) not in tossed_locs]
                    new_v_slats.append(new_v_slat)
                self.v_slats = new_v_slats

    def add_generic_glues(self, slat_group, coop, orientation, glue_marker, generic_glues=None):
        outputs = generic_glues
        for output in outputs:
            generic_label = output[2]
            keeper_function = lambda n: True
            scaled_coords = (output[0][0] * coop, output[0][1] * coop)
            tossed_locs = []
            for x in range(coop):
                x_coord = scaled_coords[0] + x
                for y in range(coop):
                    y_coord = scaled_coords[1] + y
                    self.coords_to_glues_map[(x_coord, y_coord)] = f'{generic_label}({x},{y})'

            if orientation == Orient.H:
                new_h_slats = []
                for h_slat in self.h_slats:
                    new_h_slat = [(x, y) for (x, y) in h_slat if (x, y) not in tossed_locs]
                    new_h_slats.append(new_h_slat)
                self.h_slats = new_h_slats
            else:
                new_v_slats = []
                for v_slat in self.v_slats:
                    new_v_slat = [(x, y) for (x, y) in v_slat if (x, y) not in tossed_locs]
                    new_v_slats.append(new_v_slat)
                self.v_slats = new_v_slats

    def add_gadget_portion(self, gadget_portion, glue_label, glue_strength, coop):
        overload_sets = []
        for slat_group in gadget_portion:
            if 'overload' in slat_group:
                overload_sets += slat_group['overload']
        # use only slat_groups that match overload set
        usable_slat_groups = []
        for slat_group in gadget_portion:
            if len(overload_sets) == 0:
                usable_slat_groups.append(slat_group)
            elif (not 'overload' in slat_group):
                usable_slat_groups.append(slat_group)
            elif ('overload' in slat_group) and (self.overload_set in slat_group['overload']):
                usable_slat_groups.append(slat_group)
        #print(f'overload_sets: {overload_sets}, usable_slat_groups: {usable_slat_groups}')
        for slat_group in usable_slat_groups:
            b_need_orientation = False
            if len(slat_group['coords']) == 1:
                coords_start = coords_end = slat_group['coords'][0]
                b_need_orientation = True
            else:
                coords_start = slat_group['coords'][0]
                coords_end = slat_group['coords'][1]
            if 'orientation' in slat_group:
                orientation = slat_group['orientation']
            elif b_need_orientation:
                print(f'ERROR: Orientation must be provided for gadget_portion with only one coordinate: {gadget_portion}')
                quit()
            elif is_horizontal([coords_start, coords_end]):
                orientation = Orient.H
            elif is_vertical([coords_start, coords_end]):
                orientation = Orient.V
            else:
                print(f'ERROR: Invalid slat orientation with start {coords_start} and end {coords_end}')
                quit()

            slat_coords = get_points_from_ends(coords_start, coords_end)
            slat_coords = get_scaled_slat(slat_coords, coop, orientation)

            # Keep only slats in 'active_slats' locations (which defaults to all)
            active_slats = list(range(coop))
            if 'active_slats' in slat_group:
                args = slat_group['active_slats'][1]
                active_slats_function = slat_group['active_slats'][0](args)
                active_slats = [i for i in range(coop) if active_slats_function(i)]

            for i in range(coop):
                if i in active_slats:
                    if orientation == Orient.H:
                        curr_slat = [(x,y+i) for (x,y) in slat_coords]
                        curr_slat.sort()
                        self.h_slats.append(curr_slat)
                        if 'color' in slat_group:
                            self.slat_color_map[get_slat_name_from_coord_list(curr_slat)] = slat_group['color']
                
                    else:
                        curr_slat = [(x+i,y) for (x,y) in slat_coords]
                        curr_slat.sort()
                        self.v_slats.append(curr_slat)
                        if 'color' in slat_group:
                            self.slat_color_map[get_slat_name_from_coord_list(curr_slat)] = slat_group['color']

            #TODO: determine if/how to apply motif...

            glue_marker = ''
            if 'marker' in slat_group:  # There's a marker to account for overloaded I/O
                glue_marker = slat_group['marker']

            # Account for output glues:
            #   1. Make sure that outputting slat extensions match io_conventions
            #   2. Add the necessary glues to self.coords_to_glues_map
            outputs = []
            if ('outputs' in slat_group):
                self.add_output_glues(slat_group, coop, glue_strength, glue_label, orientation, glue_marker,
                                      outputs=slat_group['outputs'])

            if ('generic_glues' in slat_group):
                self.add_generic_glues(slat_group, coop, orientation, glue_marker,
                                       generic_glues=slat_group['generic_glues'])

            # Add input glues, with suffix of V or H, opposite of slat orientation
            if 'inputs' in slat_group:
                inputs = slat_group['inputs']
                for input_def in inputs:
                    scaled_coords = (input_def[0][0]*coop, input_def[0][1]*coop)
                    if glue_strength == 2:
                        marker_function = even_and_odd
                    else:
                        marker_function = self.io_conventions((input_def[1],'in'))
                    for x in range(coop):
                        x_coord = scaled_coords[0] + x
                        for y in range(coop):
                            y_coord = scaled_coords[1] + y

                            if orientation == Orient.H:
                                if input_def[1] not in [N,S]:
                                    print(f'ERROR: Must have vertical input for horizontal slat for slat_group {slat_group}')
                                    quit()
                                elif marker_function(x):
                                    self.coords_to_glues_map[(x_coord,y_coord)] = f'{glue_label}{glue_marker}({x},{y})V'
                            else:
                                if input_def[1] not in [E,W]:
                                    print(f'ERROR: Must have horizontal input for vertical slat for slat_group {slat_group}')
                                    quit()
                                elif marker_function(y):
                                    self.coords_to_glues_map[(x_coord,y_coord)] = f'{glue_label}{glue_marker}({x},{y})H'
            
            if 'overload' in slat_group:
                pass # TODO: handle this!

            if 'empty_coords' in slat_group:
                for empty_coords in slat_group['empty_coords']:
                    for x in range(coop):
                        for y in range(coop):
                            ec = (empty_coords[0]*coop + x, empty_coords[1]*coop + y)
                            self.empty_coords.append(ec)

    def create_slats_for_tile_type(self, io_tile_type, coop):
        self.name = io_tile_type.name
        self.color = io_tile_type.color
        input_sides = get_tile_input_sides(io_tile_type)
        output_sides = get_tile_output_sides(io_tile_type)

        # Validate configuration of input/output sides. Current restrictions are that there are either:
        #   (1) no inputs (for seed tiles), or
        #   (2) input glue strengths must sum to 2 and input glues must be adjacent (i.e. no across-the-gap cooperation).
        if not len(input_sides) in [1, 2]:
            if not len(input_sides) == 0:
                print(f'ERROR: tile {io_tile_type.name} has an unsupported number of input sides. Input sides: {input_sides}')
                exit()
            else:
                print(f'Tile {io_tile_type.name} has no input sides, so is assumed to be a seed tile.')
        else:
            input_strength_sum = 0
            for input_side in input_sides:
                input_strength_sum += io_tile_type.sides[input_side].strength
            if input_strength_sum != 2:
                print(f'ERROR: tile {io_tile_type.name} has an unsupported sum of input glue strengths = {input_strength_sum} from input sides {input_sides}. Must equal 2.')
                exit()

        # rotation_count = 0
        # if len(input_sides) > 0:
        #     CW_most_input = get_CW_most_direction(set(input_sides))
        #     CW_most_input2 = CW_most_input
        #     while CW_most_input2 != S:
        #         CW_most_input2 = (CW_most_input2 + 1) % 4
        #         rotation_count += 1
        #
        # if rotation_count > 0:
        #     print(f'The tile type must be rotated {rotation_count*90} degrees CW for its most CW input to be on the south.')

        if len(input_sides) == 0: # Seed tile
            pass
        elif len(input_sides) == 1: # Tau-strength input
            pass
        elif (len(input_sides) == 2):  # Cooperative attachment
            if directions_are_adjacent(input_sides[0], input_sides[1]): # Adjacent input sides
                pass
            else: # Across-the-gap input sides
                pass # the gadget_additions_map will need to have an entry for across-the-gap for this to work
        else:
            print('ERROR: Temperature > 2 is not currently supported.')
            quit()

        # Add the base portions
            # Rotate the gadget to the orientation which matches the original io_tile_type
        #rotated_macrotile_base = rotate_gadget_definition_CW(self.macrotile_base, (4 - rotation_count)*90)
        rotated_macrotile_base = rotate_gadget_definition_CW(self.macrotile_base, self.macrotile_rotation)
        self.add_gadget_portion(rotated_macrotile_base, None, None, coop)

        # Add the portions for each side
        for d in [N,E,S,W]:
            key = get_key_for_side(d, io_tile_type)
            if key is not None:
                rot_key = ((key[0] + int(4 - self.macrotile_rotation/90)) % 4, key[1], key[2])
                if rot_key in self.macrotile_additions_map:
                    gadget_portion = self.macrotile_additions_map[rot_key]
                    rotated_gadget_portion = rotate_gadget_definition_CW(gadget_portion, self.macrotile_rotation)
                    self.add_gadget_portion(rotated_gadget_portion, io_tile_type.sides[d].label, io_tile_type.sides[d].strength, coop)
                else:
                    print(f'WARNING: rot_key {rot_key} not found in gadget definition!')
                    pass #TODO: This should be an error when we're all done...

        # Combine overlapping slats
        for slat_list in [self.v_slats, self.h_slats]:
            new_slats = []
            skip_list = []
            for i in range(len(slat_list)):
                if not i in skip_list:
                    slat1 = slat_list[i]
                    overlap_list = []
                    for j in range(i+1, len(slat_list)):
                        slat2 = slat_list[j]
                        overlaps = get_overlaps(slat1, slat2)
                        if len(overlaps) > 0:
                            overlap_list.append(slat2)
                            skip_list.append(j)

                    new_slat = copy.deepcopy(slat1)

                    slat1_color = None
                    slat1_name = get_slat_name_from_coord_list(slat1)
                    if slat1_name in self.slat_color_map:
                        slat1_color = self.slat_color_map[slat1_name]

                    overlap_slat_color = None
                    for overlap_slat in overlap_list:

                        overlap_slat_name = get_slat_name_from_coord_list(overlap_slat)
                        if overlap_slat_name in self.slat_color_map:
                            color = self.slat_color_map[overlap_slat_name]
                            if overlap_slat_color is None:
                                overlap_slat_color = color
                            elif overlap_slat_color != color:
                                print(f'ERROR: Attempting to merge two slats that have different colors: {overlap_slat} {overlap_slat_color} {color}')
                                quit()

                        new_slat = list(set(new_slat).union(set(overlap_slat)))
                    new_slat.sort()
                    new_slats.append(new_slat)

                    if (slat1_color is not None) or (overlap_slat_color is not None):
                        if (slat1_color is not None) and (overlap_slat_color is not None) and (slat1_color != overlap_slat_color):
                            print(f'ERROR: Attempting to merge two slats that have different colors: {slat1} {slat1_color} {overlap_slat_color}')
                            quit()
                        new_slat_name = get_slat_name_from_coord_list(new_slat)
                        new_slat_color = slat1_color if (slat1_color is not None) else overlap_slat_color
                        self.slat_color_map[new_slat_name] = new_slat_color

            if slat_list == self.v_slats:
                self.v_slats = new_slats
            elif slat_list == self.h_slats:
                self.h_slats = new_slats
            else:
                print(f'ERROR: Failed to combine overlapping slats!')
                quit()

        # Find the set of locations for 'internal' domains that bind vertical slats of the gadget to horizontal slats of the same gadget
        self.crossing_coords = []
        for v_slat in self.v_slats:
            for h_slat in self.h_slats:
                for h_coord in h_slat:
                    if h_coord in v_slat:
                        if not h_coord in self.empty_coords:
                            self.crossing_coords.append(h_coord)
                        #else:
                        #    print(f'skipped empty coord {h_coord}')
        
    def getLibSlatGroup(self):
        h_slat_group = SlatGroup()
        h_count = 0
        overload_tag = '' if self.overload_set is None else f'OL{self.overload_set}'
        for h_slat in self.h_slats:
            name = f'{self.name}h{h_count}{overload_tag}{self.slat_name_suffix}'
            h_count += 1
            length = get_length(h_slat)
            h_lib_slat = LibSlat(name, h_slat[0], True, length)
            self.add_glues_to_lib_slat(h_lib_slat, h_slat, self.name)
            h_slat_name = get_slat_name_from_coord_list(h_slat)
            if h_slat_name in self.slat_color_map:
                h_lib_slat.color = self.slat_color_map[h_slat_name]
            else:
                h_lib_slat.color = self.color
            h_slat_group.slats.append(h_lib_slat)
            
        v_slat_group = SlatGroup()
        v_count = 0
        for v_slat in self.v_slats:
            name = f'{self.name}v{v_count}{overload_tag}{self.slat_name_suffix}'
            v_count += 1
            length = get_length(v_slat)
            v_lib_slat = LibSlat(name, v_slat[0], False, length)
            self.add_glues_to_lib_slat(v_lib_slat, v_slat, self.name)
            v_slat_name = get_slat_name_from_coord_list(v_slat)
            if v_slat_name in self.slat_color_map:
                v_lib_slat.color = self.slat_color_map[v_slat_name]
            else:
                v_lib_slat.color = self.color
            v_slat_group.slats.append(v_lib_slat)

        macrotile_slat_group = GadgetSlatGroup(gadget_name = self.name, horizontal_slats = h_slat_group, vertical_slats = v_slat_group)
        return macrotile_slat_group

def get_overload_map_for_signature(macrotile_additions_map, signature):
    overload_map = {}
    for d in [N,E,S,W]:
        key = (d, signature[d][0], signature[d][1])
        if key in macrotile_additions_map:
            slats_list = macrotile_additions_map[key]
            for slats in slats_list:
                if 'overload' in slats:
                    overloads = slats['overload']
                    for overload in overloads:
                        if overload in overload_map:
                            overload_map[overload].append(key)
                        else:
                            overload_map[overload] = [key]
    #if len(overload_map) > 0:
    #    print(f'signature {signature} has overload_map {overload_map}')
    return overload_map

def get_slats_from_tile_system(macrotile_def, io_tile_system, scale_factor, motif):
    macrotile_slat_groups = []
    tile_to_slat_dict = {}

    for io_tile_type_name in io_tile_system.tileTypes:
        print(f'Creating slats for tile type {io_tile_type_name}')
        io_tile_type = io_tile_system.tileTypes[io_tile_type_name]
        #slats = get_slats_for_io_tile_type(io_tile_type, scale_factor, motif)

        signature = get_io_tile_type_signature(io_tile_type)
        matching_map_list = get_maps_for_signature(signature, macrotile_def.macrotile_mapping_list)
        if len(matching_map_list) == 0:
            print(f'ERROR: There is no matching map for IO tile type {io_tile_type_name} with signature {signature}.')
            quit()
        else:
            map_names = ''
            for matching_map in matching_map_list:
                map_rotation = matching_map['rotation']
                if 'name' in matching_map['macrotile_additions_map']:
                    map_names += f"{matching_map['macrotile_additions_map']['name']} (rot {map_rotation}) "
                else:
                    map_names += "<no name> (rot {map_rotation}) "
            #print(f"  Its signature {signature} matches the following macrotile_additions_maps: {map_names}")

        # Make a macrotile for every matching map
        map_num = 1
        slat_name_suffix = ''
        for matching_map in matching_map_list:
            if len(matching_map_list) > 1:
                slat_name_suffix = f'_map{map_num}'
                map_num += 1
            overload_map = get_overload_map_for_signature(matching_map['macrotile_additions_map'], signature)
            overload_keys = list(overload_map.keys())
            overload_keys.sort()
            overload_sets = [None] + list(overload_keys)
            if len(overload_sets) > 1:
                print(f"IO tile type '{io_tile_type_name}' with signature '{signature}' has overload sets {overload_sets[1:]}")
            slat_groups = []
            for overload_set in overload_sets:
                macrotile = Macrotile(scale_factor, matching_map['macrotile_base'], matching_map['macrotile_additions_map'],
                                      matching_map['rotation'], macrotile_def.h_glue_suffix, macrotile_def.v_glue_suffix,
                                      macrotile_def.io_conventions, overload_set, slat_name_suffix, motif)

                macrotile.create_slats_for_tile_type(io_tile_type, scale_factor)
                if (len(macrotile.v_slats) > 0) or (len(macrotile.h_slats) > 0):
                    slat_group = macrotile.getLibSlatGroup()
                    slat_groups.append(slat_group)

            # Keep the first macrotile, then remove slats from the others if they were in the first
            kept_slat_groups = [slat_groups[0]]
            current_slats = copy.deepcopy(slat_groups[0].horizontal_slat_group.slats)
            current_slats += copy.deepcopy(slat_groups[0].vertical_slat_group.slats)
            for slat_group in slat_groups[1:]:
                pruned_slat_group = copy.deepcopy(slat_group)
                pruned_slat_group.horizontal_slat_group = SlatGroup()
                pruned_slat_group.vertical_slat_group = SlatGroup()
                b_keep = False
                for slat in slat_group.horizontal_slat_group.slats:
                    b_found = False
                    for curr_slat in current_slats:
                        if slat.equals(curr_slat):
                            b_found = True
                            #print(f'found a duplicate! {slat}')
                    if not b_found:
                        pruned_slat_group.horizontal_slat_group.slats.append(slat)
                        b_keep = True
                        current_slats.append(slat)
                for slat in slat_group.vertical_slat_group.slats:
                    b_found = False
                    for curr_slat in current_slats:
                        if slat.equals(curr_slat):
                            b_found = True
                            #print(f'found a duplicate! {slat}')
                    if not b_found:
                        pruned_slat_group.vertical_slat_group.slats.append(slat)
                        b_keep = True
                        current_slats.append(slat)
                if b_keep:
                    kept_slat_groups.append(pruned_slat_group)

            for slat_group in kept_slat_groups:
                if io_tile_type.name not in tile_to_slat_dict:    #TODO: How should we handle multiple macrotiles for one tile type??
                    tile_to_slat_dict[io_tile_type.name] = slat_group
                macrotile_slat_groups.append(slat_group)

    seed_slat_groups = []
    for seed_entry in io_tile_system.seedPatch:
        #TODO: Need to determine if there are glues between seed tiles and if so, make one of each pair an input and the other output
        seed_tile_slats = tile_to_slat_dict[seed_entry[0]]
        x_addition = (3, -3)
        y_addition = (3, 3)
        x_loc = (seed_entry[1][0] * x_addition[0] * scale_factor) + (seed_entry[1][1] * y_addition[0] * scale_factor)
        y_loc = (seed_entry[1][0] * x_addition[1] * scale_factor) + (seed_entry[1][1] * y_addition[1] * scale_factor)
        seed_tile_slats.translate((x_loc, y_loc))
        seed_slat_groups.append(seed_tile_slats)

    slat_num = 0
    for slat_group in macrotile_slat_groups:
        slat_num = slat_num + len(slat_group.horizontal_slat_group.slats)
        slat_num = slat_num + len(slat_group.vertical_slat_group.slats)

    seed_slat_num = 0
    for slat_group in seed_slat_groups:
        seed_slat_num = seed_slat_num + len(slat_group.horizontal_slat_group.slats)
        seed_slat_num = seed_slat_num + len(slat_group.vertical_slat_group.slats)

        # Ensure that slats in the seed-specific gadgets are marked as only being allowed in the seed assembly
        for slat in slat_group.horizontal_slat_group.slats:
            slat.b_seed_only = True
        for slat in slat_group.vertical_slat_group.slats:
            slat.b_seed_only = True

    print(f'Slat set contains {slat_num+seed_slat_num} slat types. ({seed_slat_num} are seed slat types, {slat_num} are regular slat types.)')
    return macrotile_slat_groups, seed_slat_groups


##################################################################
# IO tile helper functions

def get_io_tile_type_signature(io_tile_type):
    if not tile_type_is_IO_marked(io_tile_type):
        print(f'ERROR: Tile type {io_tile_type} is not an IO marked tile type')
        quit()
    signature = {}
    for direction in [N,E,S,W]:
        side_sig = get_signature_for_side(direction, io_tile_type)
        signature[direction] = side_sig
    return signature

def io_tile_type_signature_matches(signature, signature_mapping):
    #print(f'signature: {signature}, map: {signature_mapping}')
    #Examples:
    #signature: {0: ('out', 1), 1: ('out', 2), 2: (None, 0), 3: (None, 0)}
    #signature_mapping: {'in': {S:2}, 'out': {W:[0,1,2], N:[0,1,2], E:[0,1,2]}
    for io_type in ['in', 'out']:
        io_map = signature_mapping[io_type]
        for d in io_map:
            if type(io_map[d]) is list:
                strength_list = io_map[d]
            else:
                strength_list = [io_map[d]]
            if not signature[d][1] in strength_list:
                return False
            elif (signature[d][0] != io_type) and not ((signature[d][0] == None) and (0 in strength_list)):
                return False
    return True

def get_maps_for_signature(signature, macrotile_mapping_list):
    matching_maps = []
    for macrotile_mapping in macrotile_mapping_list:
        b_added = False
        for signature_mapping in macrotile_mapping['signatures']:
            if (not b_added) and io_tile_type_signature_matches(signature, signature_mapping):
                matching_maps += macrotile_mapping['maps']
                b_added = True
    return matching_maps

def fix_reverse_IO_markings(tile_type_dict):
    for tile_type_name in tile_type_dict:
        tile_type = tile_type_dict[tile_type_name]
        for side_dir in tile_type.sides:
            glue = tile_type.sides[side_dir]
            if glue.label.endswith(INPUT_MARKERS[side_dir]):
                glue.label = f'{INPUT_MARKERS[side_dir]}{glue.label[:-1]}'
            elif glue.label.endswith(OUTPUT_MARKERS[side_dir]):
                glue.label = f'{OUTPUT_MARKERS[side_dir]}{glue.label[:-1]}'

def tile_type_is_IO_marked(tile_type):
    for side_dir in tile_type.sides:
        glue = tile_type.sides[side_dir]
        if glue.strength > 2:
            print(f'ERROR: Only temperature 2 systems, with maximum glue strength 2, are supported.')
            exit()
        elif ((glue.strength > 0) and (not glue.label.startswith(INPUT_MARKERS[side_dir])) and
              (not glue.label.startswith(OUTPUT_MARKERS[side_dir]))):
            return False
    return True

def tile_types_are_IO_marked(tile_type_dict):
    for tile_type_name in tile_type_dict:
        tile_type = tile_type_dict[tile_type_name]
        tt = tile_type_is_IO_marked(tile_type)
        if not tt:
            return False
    return True

def get_IO_tile_types(tile_type, b_for_seed=False):
    strength1_sides = []
    strength2_sides = []
    b_io_marked = True
    for side_dir in tile_type.sides:
        glue = tile_type.sides[side_dir]
        if glue.strength == 1:
            strength1_sides.append(side_dir)
        elif glue.strength == 2:
            strength2_sides.append(side_dir)
        elif glue.strength > 2:
            print(f'ERROR: Only temperature 2 systems, with maximum glue strength 2, are supported.')
            exit()
        if (glue.strength > 0) and (not glue.label.startswith(INPUT_MARKERS[side_dir])) and (not glue.label.startswith(OUTPUT_MARKERS[side_dir])):
            b_io_marked = False
    if b_io_marked:
        return {tile_type.name: tile_type}
    else:
        if b_for_seed:
            io_tile_type = tile_type.Copy()
            for side_dir in io_tile_type.sides:
                if io_tile_type.sides[side_dir].strength > 0:
                    io_tile_type.sides[side_dir].label = f'{OUTPUT_MARKERS[side_dir]}{io_tile_type.sides[side_dir].label}'
            return {io_tile_type.name: io_tile_type}
        else:
            input_sets = []
            for strength2_side in strength2_sides:
                input_sets.append([strength2_side])
            for strength1_side1 in strength1_sides:
                for strength1_side2 in strength1_sides:
                    if directions_are_adjacent(strength1_side1, strength1_side2):
                        input_sets.append((strength1_side1, strength1_side2))
            io_tile_type_dict = {}
            for input_set in input_sets:
                io_tile_type = tile_type.Copy()
                name_addition = ''
                for side_dir in io_tile_type.sides:
                    if (side_dir == input_set[0]) or ((len(input_set) == 2) and (side_dir in input_set)):
                        name_addition += DIR_TO_STR[side_dir]
                        io_tile_type.sides[side_dir].label =  f'{INPUT_MARKERS[side_dir]}{io_tile_type.sides[side_dir].label}'
                    elif io_tile_type.sides[side_dir].strength > 0:
                        io_tile_type.sides[side_dir].label = f'{OUTPUT_MARKERS[side_dir]}{io_tile_type.sides[side_dir].label}'
                io_tile_type.name = io_tile_type.name + name_addition
                io_tile_type_dict[io_tile_type.name] = io_tile_type
        return io_tile_type_dict
    

##################################################################

class Macrotile_Def:
    def __init__(self, macrotile_def_file):
        # Read the file defining the macrotile and execute the code
        file_pointer = open(macrotile_def_file, 'r')
        file_contents = file_pointer.read()

        helper_dict = {'Orient': Orient, 'N': N, 'E': E, 'S': S, 'W': W, 'U': U, 'D': D, 'odd': odd, 'even': even}
        macrotile_def_dict = {}
        exec(file_contents, helper_dict, macrotile_def_dict)

        self.h_glue_suffix = macrotile_def_dict['h_glue_suffix']
        self.v_glue_suffix = macrotile_def_dict['v_glue_suffix']
        self.io_conventions = macrotile_def_dict['io_conventions']
        #self.macrotile_base = macrotile_def_dict['macrotile_base']
        #self.macrotile_additions_map = macrotile_def_dict['macrotile_additions_map']
        self.macrotile_mapping_list = macrotile_def_dict['macrotile_mapping_list']

        # self.macrotile_base_dict = {}
        # self.macrotile_additions_map_dict = {}
        # for macrotile_mapping in macrotile_mapping_list:
        #     macrotile_base = macrotile_mapping['macrotile_base']
        #     macrotile_additions_map = macrotile_mapping['macrotile_additions_map']

##################################################################


def get_signature_inputs(signature):
    inputs = {}
    for d in [N,E,S,W]:
        if signature[d][0] == 'in':
            inputs[d] = signature[d]
    return inputs

def get_signature_outputs(signature):
    outputs = {}
    for d in [N,E,S,W]:
        if signature[d][0] == 'out':
            outputs[d] = signature[d]
    return outputs

def signature_is_in_map(signature, macrotile_signatures):
    #matched_macrotile_signatures = []
    tile_inputs = get_signature_inputs(signature)
    tile_outputs = get_signature_outputs(signature)
    for macrotile_signature in macrotile_signatures:
        macrotile_signature_input_list = copy.deepcopy(macrotile_signature['in'])
        macrotile_signature_output_list = copy.deepcopy(macrotile_signature['out'])

        for d in [N,E,S,W]:
            if d in macrotile_signature_input_list:
                if not (type(macrotile_signature_input_list[d]) is list):
                    macrotile_signature_input_list[d] = [macrotile_signature_input_list[d]]
            if d in macrotile_signature_output_list:
                if not (type(macrotile_signature_output_list[d]) is list):
                    macrotile_signature_output_list[d] = [macrotile_signature_output_list[d]]

        b_keep_going = True
        for d in [N,E,S,W]:
            b_is_zero = True
            if (d in tile_inputs) and (tile_inputs[d][1] != 0):
                b_is_zero = False
            if (d in tile_outputs) and (tile_outputs[d][1] !=0):
                b_is_zero = False
            if b_is_zero:
                if (d in macrotile_signature_input_list) and (0 not in macrotile_signature_input_list[d]):
                    b_keep_going = False
                if (d in macrotile_signature_output_list) and (0 not in macrotile_signature_output_list[d]):
                    b_keep_going = False
            elif d in tile_inputs:
                if d not in macrotile_signature_input_list:
                    b_keep_going = False
                elif tile_inputs[d][1] not in macrotile_signature_input_list[d]:
                    b_keep_going = False
            elif d in tile_outputs:
                if d not in macrotile_signature_output_list:
                    b_keep_going = False
                elif tile_outputs[d][1] not in macrotile_signature_output_list[d]:
                    b_keep_going = False
            else:
                print(f"ERROR: Invalid case encountered in 'signature_is_in_map'")
                quit()
        if b_keep_going:
            return True
    return False

def generate_io_tile_set_for_map(macrotile_def):
    io_combos = [(None, 0), ('in', 1), ('in', 2), ('out', 1), ('out', 2)]
    matched_signatures = []
    for io_combo_S in io_combos:
        for io_combo_W in io_combos:
            for io_combo_N in io_combos:
                for io_combo_E in io_combos:
                    # input_sum = io_combo_S[1] if io_combo_S[0] == 'in' else 0
                    # input_sum += io_combo_W[1] if io_combo_W[0] == 'in' else 0
                    # input_sum += io_combo_N[1] if io_combo_N[0] == 'in' else 0
                    # input_sum += io_combo_E[1] if io_combo_E[0] == 'in' else 0
                    # if input_sum > 0:
                    
                    signature = {N:io_combo_N, E:io_combo_E, S:io_combo_S, W:io_combo_W}
                    for macrotile_mapping in macrotile_def.macrotile_mapping_list:
                        if signature_is_in_map(signature, macrotile_mapping['signatures']):
                            if (io_combo_N[1] > 0) or (io_combo_E[1] > 0) or (io_combo_S[1] > 0) or (io_combo_W[1] > 0):    # Make sure there is at least one glue
                                matched_signatures.append(signature)
                            else:
                                print(f"WARNING: The following macrotile_mapping matches a tile type with no glues: '{macrotile_mapping}'")
    io_tile_set = {}
    print(f'Number of matched signatures: {len(matched_signatures)}')
    for signature in matched_signatures:
        #print(f'matched signature: {signature}')
        tt_name = str(signature).replace(' ','').replace("'","").replace("{","").replace("}","")
        io_tile_type = PyTASTileType(name=tt_name, label=tt_name)
        input_set = set()
        for d in [N,E,S,W]:
            if (d in signature) and (not signature[d][0] is None):
                sig_d = signature[d]
                if sig_d[0] == 'in':
                    io_tile_type.sides[d] = Glue(sig_d[1], f'{INPUT_MARKERS[d]}{DIR_TO_STR[d]}')
                    input_set.add(d)
                else:
                    io_tile_type.sides[d] = Glue(sig_d[1], f'{OUTPUT_MARKERS[d]}{DIR_TO_STR[d]}')
        io_tile_set[io_tile_type.name] = io_tile_type
        if len(input_set) == 0:
            io_tile_type.color = 'Green'
        else:
            cw_d = get_CW_most_direction(input_set)
            colors = ['Blue', 'Yellow', 'Red', 'White']
            io_tile_type.color = colors[cw_d]
    #for iott in io_tile_set:
    #    print(f'io tile type {io_tile_set[iott]}')
    return io_tile_set

def output_tiles_to_PDF_or_SVG(tile_set, base_file_name):
    pdf_file_name = None
    svg_file_name = None
    if b_generate_pdf:
        pdf_file_name = f'{base_file_name}.pdf'
    if b_generate_svg:
        svg_file_name = f'{base_file_name}.svg'

    num_tile_types = len(tile_set)
    num_tiles_per_row = math.floor(math.sqrt(num_tile_types))
    num_rows = math.ceil(num_tile_types / num_tiles_per_row)
    pageWidth = num_tiles_per_row * CELLSIZE
    pageHeight = num_rows * CELLSIZE

    dwg = None
    pdf_canvas = None
    if not svg_file_name is None:
        dwg = svgwrite.Drawing(filename=svg_file_name, debug=True)
        dwg.viewbox(width=pageWidth, height=pageHeight)
    if not pdf_file_name is None:
        pdf_canvas = canvas.Canvas(pdf_file_name, pagesize=(pageWidth, pageHeight))

    tile_count = 0
    body_dims=CELLSIZE
    for tile_type_name in tile_set:
        spacing = 5
        x = (tile_count % num_tiles_per_row) * (body_dims + spacing)
        y = int(tile_count / num_tiles_per_row) * (body_dims + spacing)
        tile_type = tile_set[tile_type_name]
        tile_type.MakeImage(pdfCanvas=pdf_canvas, svgCanvas=dwg, svg_trans=(x, y), pdf_trans=(x, y))
        tile_count += 1

    if not svg_file_name is None:
        dwg.save()
        print(f"Tiles saved to '{svg_file_name}'.")
    if not pdf_file_name is None:
        pdf_canvas.save()
        print(f"Tiles saved to '{pdf_file_name}'.")

def get_flattened_slat_lists(macrotile_slat_groups):
    vertical_slats = []
    horizontal_slats = []
    for macrotile_slat_group in macrotile_slat_groups:
        for slat in macrotile_slat_group.vertical_slat_group.slats:
            vertical_slats.append(slat)
        for slat in macrotile_slat_group.horizontal_slat_group.slats:
            horizontal_slats.append(slat)
    return vertical_slats, horizontal_slats

def remove_duplicate_slats(slat_list):
    new_slat_list = []
    for i in range(len(slat_list)):
        b_found = False
        for j in range(i+1, len(slat_list)):
            if slat_list[i].equals(slat_list[j]):
                b_found = True
        if not b_found:
            new_slat_list.append(slat_list[i])
    return new_slat_list

def create_slat_system_from_aTAM_system(input_macrotile_def_file, tdp_file_name, output_directory='.', base_file_name=None,
                                        scale_factor=4, slat_count=None, motif=None, color_scheme='tile', b_generate_xml=True,
                                        b_generate_pdf=False, b_generate_svg=False, b_show_labels=True, b_convert_full_map=False):
    xml_file_name = None
    pdf_file_name = None
    svg_file_name = None
    b_add_flipped_slats = False

    macrotile_def = Macrotile_Def(input_macrotile_def_file)

    if b_convert_full_map:
        tdp_file_name = f'{input_macrotile_def_file[:-3]}_full'
    elif tdp_file_name is None:
        print(f'ERROR: a tdp file must be specified as input\n')
        quit()
    else:
        tdp_file_name = os.path.abspath(tdp_file_name)
        if output_directory is None:
            output_directory = os.path.dirname(tdp_file_name)

    if output_directory is None:
        output_directory = "."
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    # Get just the file name
    tdp_base_file_name = os.path.basename(tdp_file_name)

    # Read the tile assembly system
    tile_system = PyTASTileSystem()

    if b_convert_full_map:
        tile_set = generate_io_tile_set_for_map(macrotile_def)
        tile_system.tileTypes = tile_set
        if b_generate_pdf or b_generate_svg:
            output_tiles_to_PDF_or_SVG(tile_set, f'{output_directory}/{tdp_base_file_name}_tiles')
    else:
        tile_system.LoadFromFile(tdp_file_name)
    
    fix_reverse_IO_markings(tile_system.tileTypes)

    # Determine if the system's tile types are already marked as IO
    if tile_types_are_IO_marked(tile_system.tileTypes):
        io_tile_system = tile_system
    else:
        io_tile_system = PyTASTileSystem()
        io_tile_system.temperature = tile_system.temperature

        io_tile_type_dict = {}
        for tile_type_name in tile_system.tileTypes:
            print(f'Creating IO tiles for tile type {tile_type_name}')

            tile_type = tile_system.tileTypes[tile_type_name]
            if len(tile_type.sides) > 4:
                print(f'ERROR: This conversion only works for 4-sided tiles.')
                exit()

            # If tile type is in the seed, make a copy of it with no input sides
            b_in_seed = False
            for seed_entry in tile_system.seedPatch:
                if tile_type_name == seed_entry[0]:
                    b_in_seed = True
                    partial_io_tile_type_dict = get_IO_tile_types(tile_type, True)
                    io_tile_type_dict.update(partial_io_tile_type_dict)

            # Could only make 'seed' versions of tiles in seed if desired...
            if True: #not b_in_seed:
                partial_io_tile_type_dict = get_IO_tile_types(tile_type)
                io_tile_type_dict.update(partial_io_tile_type_dict)

        io_tile_system.tileTypes = io_tile_type_dict
        io_tile_system.seedPatch = copy.deepcopy(tile_system.seedPatch)

        # Output the IO marked system (tdp and tds files)
        io_system_base_file_name = f'{output_directory}/{tdp_base_file_name[:-4]}_IO'
        io_tile_system.SaveToFile(io_system_base_file_name)
        print(f'Saving newly generated IO tile system to {io_system_base_file_name}')

    # Make sure that there are not multiple tiles with the same input set
    input_side_set_dict = {}
    for io_tile_type_name in io_tile_system.tileTypes:
        io_tile_type = io_tile_system.tileTypes[io_tile_type_name]
        input_side_set = []
        for side_dir in io_tile_type.sides:
            glue = io_tile_type.sides[side_dir]
            if (glue.strength > 0) and glue.label.startswith(INPUT_MARKERS[side_dir]):
                input_side_set.append((side_dir, glue.strength, glue.label))
        #print(f'input_side_set: {input_side_set}')
        if len(input_side_set) > 0:
            for io_ttype_name in input_side_set_dict:
                iss = input_side_set_dict[io_ttype_name]
                if (iss == input_side_set) and (not b_convert_full_map):
                    print(f'WARNING: multiple IO tiles have the same input set of {iss}: {io_ttype_name} and {io_tile_type_name}')
                    #exit()
            input_side_set_dict[io_tile_type_name] = input_side_set

    if io_tile_system.temperature != 2:
        print(f'ERROR: This conversion only works for temperature=2 systems currently. The input system has temperature={tile_system.temperature}')
        exit()

    if (motif is not None) and (len(motif) != scale_factor):
        print(f'ERROR: motif length must equal the scale_factor\n')
        DisplayHelp()

    if slat_count is None:
        if not motif is None:
            slat_count = len(list(set(motif)))
        else:
            slat_count = scale_factor

    if motif is None:
        motif_chars = []
        for i in range(slat_count):
            motif_chars.append(chr(ord('a') + i))
        motif = []
        for i in range(scale_factor):
            motif.append(motif_chars[i % len(motif_chars)])
    motif_str = ''.join(map(str, motif))
    print(f"Using motif '{motif_str}'")

    extend_string = ''
    # if not extended_slat_length is None:
    #     extend_string = f'-extend{extended_slat_length}'

    if base_file_name is None:
        base_file_name = f'{output_directory}/{tdp_base_file_name[:-4]}-scale{scale_factor}-dup{slat_count}{extend_string}'
        if not motif is None:
            base_file_name += f'-motif({motif_str})'

    # tile_set = []
    #
    # tile_type = TileType(name='seed1', label='testSeed1')
    # tile_type.sides = {N: Glue(1, '^n'), E: Glue(1, '>e'), S: Glue(1, 'vs'), W: Glue(1, '<w')}
    # tile_set.append(tile_type)

    macrotile_slat_groups, seed_slat_groups = get_slats_from_tile_system(macrotile_def, io_tile_system, scale_factor, motif)

    # Check for and discard functionally duplicate slats
    all_vertical_slats, all_horizontal_slats = get_flattened_slat_lists(macrotile_slat_groups)
    new_vertical_slats = remove_duplicate_slats(all_vertical_slats)
    if len(all_vertical_slats) != len(new_vertical_slats):
        print(f' Removed {len(all_vertical_slats) - len(new_vertical_slats)} functionally duplicate vertical slats.')
    new_horizontal_slats = remove_duplicate_slats(all_horizontal_slats)
    if len(all_horizontal_slats) != len(new_horizontal_slats):
        print(f' Removed {len(all_horizontal_slats) - len(new_horizontal_slats)} functionally duplicate horizontal slats.')

    slat_system = SlatSystem()
    slat_system.b_use_rotation = False
    slat_system.binding_threshold = scale_factor
    for slats in [new_vertical_slats, new_horizontal_slats]:
        for slat in slats:
            #if (color_scheme == 'motif') and (not motif is None) and (slat.name[-1] in motif):
                #SetSlatTypeToMotifColor(None, slat, motif)
            tas_slat_type = slat_system.GetPolyominoTypeByName(slat.name)
            if tas_slat_type is None:
                tas_slat_type = GetTASSlatTypeFromLibSlatType(slat)
                #if (color_scheme == 'motif') and (not motif is None) and (tas_slat_type.name[-1] in motif):
                    #SetSlatTypeToMotifColor(tas_slat_type, None, motif)
                slat_system.AddPolyominoType(tas_slat_type)
                if b_add_flipped_slats == True:
                    rotated_slat = tas_slat_type.GetRotatedVersion(180)
                    slat_system.AddPolyominoType(rotated_slat)
            else:
                print(f"ERROR: The system already has a slat with name {slat.name}, but they are not equivalent")
                quit()
                    
    flip_text = '' if b_add_flipped_slats == False else ' (including flipped copies)'
    print(f'Slat system has {len(slat_system.PolyominoType_dict)} non-seed slat types{flip_text}')

    #domain_set = GetDomainSetFromSlatSystem(slat_system)
    #print(f'Slat system has {len(domain_set)} unique domains (a.k.a. handles) not including seed-specific slats:')

    # Check for and discard functionally duplicate slats
    all_vertical_seed_slats, all_horizontal_seed_slats = get_flattened_slat_lists(seed_slat_groups)
    new_vertical_seed_slats = remove_duplicate_slats(all_vertical_seed_slats)
    if len(all_vertical_seed_slats) != len(new_vertical_seed_slats):
        print(f' Removed {len(all_vertical_seed_slats) - len(new_vertical_seed_slats)} functionally duplicate vertical seed slats.')
    new_horizontal_seed_slats = remove_duplicate_slats(all_horizontal_seed_slats)
    if len(all_horizontal_seed_slats) != len(new_horizontal_seed_slats):
        print(f' Removed {len(all_horizontal_seed_slats) - len(new_horizontal_seed_slats)} functionally duplicate horizontal seed slats.')

    for slats in [new_vertical_seed_slats, new_horizontal_seed_slats]:
        for slat in slats:
            #if (color_scheme == 'motif') and (not motif is None) and (slat.name[-1] in motif):
                #SetSlatTypeToMotifColor(None, slat, motif)
            tas_seed_slat_type = slat_system.GetPolyominoTypeByName(slat.name)
            if tas_seed_slat_type is None:
                tas_seed_slat_type = GetTASSlatTypeFromLibSlatType(slat, b_seed_only=True)
                #if (color_scheme == 'motif') and (not motif is None) and (tas_seed_slat_type.name[-1] in motif):
                    #SetSlatTypeToMotifColor(tas_seed_slat_type, None, motif)
                slat_system.AddPolyominoType(tas_seed_slat_type)
            #else: ## This is not an error, because the slat type should have already been added to the system from the macrotile_slat_groups
            #    print(f'ERROR: The system already has a seed slat with name {slat.name} but they are not equivalent.')
            #    #quit()
            tas_slat = TASSlat(tas_seed_slat_type)
            z_coord = 1 if slat.b_horizontal else 0
            translation_coords = (slat.begin[0], slat.begin[1], z_coord)
            tas_slat.translate(translation_coords)
            slat_system.seed.AddPolyominoForViewingOnly(tas_slat)
            #print(f'added slat {tas_slat}')

    if b_generate_xml:
        xml_file_name = f'{base_file_name}.xml'
        slat_system.WriteToFile(xml_file_name, color_mapping=wxPython_color_mapping)
        print(f'Output slat system to {xml_file_name}')

    if b_generate_pdf or b_generate_svg:
        if b_generate_pdf:
            pdf_file_name = f'{base_file_name}.pdf'
        if b_generate_svg:
            svg_file_name = f'{base_file_name}.svg'

        DrawSlatsToFiles(macrotile_slat_groups, color_mapping={}, svg_file_name=svg_file_name,
                         pdf_file_name=pdf_file_name, b_show_labels=True, tile_set=io_tile_system.tileTypes)
        
    return base_file_name


#######################################################################################################################################################
## Main
if __name__ == "__main__":
    try:
        # starts at the second element of argv since the first one is the script name
        # extraparams are extra arguments passed after all option/keywords are assigned
        # opts is a list containing the pair "option"/"value"
        opts, extraparams = getopt.getopt(sys.argv[1:], "ht:o:d:n:c:m:r:l:i:xspf",
                            ["help", "input=", "tdp=", "out=", "dir=", "scale=", "count=", "motif=", "color=", "labels=", "xml", "svg", "pdf", "full"])
        #print 'Opts:',opts
        #print 'Extra parameters:',extraparams
    except:
        print("")
        print("Unrecognized option:" + str(sys.argv[1:]))
        DisplayHelp()

    if len(opts) == 0:
        DisplayHelp()

    # Add additional color settings
 #   extra_slat_color_mapping = {}
 #   slat_color_mapping.update(extra_slat_color_mapping)

    scale_factor = 4
    slat_count = None
    input_macrotile_def_file = None
    tdp_file_name = None
    b_convert_full_map = False
    output_directory = None
    base_file_name = None
    b_generate_xml = False
    b_generate_pdf = False
    b_generate_svg = False
    b_add_flipped_slats = False
    motif_str = None
    motif = None
    extended_slat_length = None
    color_scheme = 'tile'
    b_show_labels = True

    # ht:o:d:n:c:m:r:l:xsp
    #["help", "tdp=", "out=", "dir=", "scale=", "count=", "motif=", "color=", "labels=", "xml", "svg", "pdf"]
    for o,p in opts:
        if o in ['-h', '--help']:
            DisplayHelp()
        elif o in ['-i', '--input']:
            input_macrotile_def_file = p
            print(f'Macrotile definition file: {input_macrotile_def_file}')
        elif o in ['-t', '--tdp']:
            tdp_file_name = p
            print(f'tdp_file_name: {tdp_file_name}')
        elif o in ['-f', '--full']:
            b_convert_full_map = True
        elif o in ['-o', '--out']:
            base_file_name = p
        elif o in ['-d', '--dir']:
            output_directory = p
        elif o in ['-n', '--scale']:
            scale_factor = int(p)
        elif o in ['-c', '--count']:
            slat_count = int(p)
        elif o in ['-m', '--motif']:
            motif_str = p
            motif = list(p)
        elif o in ['-r', '--color']:
            if p in ['m', 'M']:
                color_scheme = 'motif'
        elif o in ['-l', '--labels']:
            if (p == 'F') or (p == 'False'):
                b_show_labels = False
        elif o in ['-x', '--xml']:
            b_generate_xml = True
        elif o in ['-s', '--svg']:
            b_generate_svg = True
        elif o in ['-p', '--pdf']:
            b_generate_pdf = True
        # elif o in ['-f', '--flip']:
        #     b_add_flipped_slats = True if p in ['T','t','True','true'] else False
        # elif o in ['-x', '--extend']:
        #     extended_slat_length = int(p)
        else:
            DisplayHelp()

    if input_macrotile_def_file is None:
        print(f'\nERROR: a Python file with definitions for the macrotiles be specified as input.')
        DisplayHelp()

    # Read the file defining the macrotile and execute the code
    #file_pointer = open(input_macrotile_def_file, 'r')
    #file_contents = file_pointer.read()
    #exec(file_contents)

    #print(f'file_contents: {file_contents}')
    #print(f'h_glue_suffix: {h_glue_suffix}')
    #print(f'macrotile_base: {macrotile_base}')

    if (b_convert_full_map == False) and (tdp_file_name is None):
        print(f'\nERROR: a tdp file must be specified as input.')
        DisplayHelp()

    create_slat_system_from_aTAM_system(input_macrotile_def_file, tdp_file_name, output_directory, base_file_name,
                                        scale_factor, slat_count, motif, color_scheme, b_generate_xml, b_generate_pdf,
                                        b_generate_svg, b_show_labels, b_convert_full_map)
